///////////////////////////////////////////////////////////////////
// CExecEnv.cpp
// 
// Execution Environment
// Classes for Attributes and Function Tables
///////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "GAPS2.h"
#include "CExecEnv.h"
#include "CGPProgram.h"
#include ".\MTRandom\randomc.h"

extern TRandomMersenne RandomGen;
#define TOTALBUILTINFUNCTIONS 32
CString AllBuiltInStrings[TOTALBUILTINFUNCTIONS] = 
{
	"Map", "If", "If-Else", "While", "DoUntil", "Push", "Pop", "Replace", "Declare", ":=",
	"+", "-", "/", "*", "%", 
	"sin", "cos", "tan", "log", "sqrt", "round", "trunc", "abs",
	"not", "&&", "||", 	">", "<", ">=", "<=", "!=", "="
};

CAttributeTable::CAttributeTable()
{
	AttribTable = NULL;
	LastInList = NULL;
	TotalAttributes = 0;
}

CAttributeTable::~CAttributeTable()
{
	ClearAll();
}	

void CAttributeTable::ClearAll()
{
	struct ATTRIBUTE *DelNode;
	struct ATTRIBUTE *CurrNode;

	for (CurrNode = AttribTable; CurrNode != NULL;)
	{
		DelNode = CurrNode;
		CurrNode = CurrNode->next;

		delete DelNode;
	}

	AttribTable = NULL;
	LastInList = NULL;
	TotalAttributes = 0;
}

void CAttributeTable::ResetValues()
{
	struct ATTRIBUTE *CurrNode;

	for (CurrNode = AttribTable; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		CurrNode->Value = CString("");
	}
}


bool CAttributeTable::AddAttribute (RETURNTYPE Type, CString Name, int Level)
{
	CString BuildName;

	if (Level > 999) // Limitation
		return FALSE;

	BuildName.Format("%s.%03d", Name, Level);

	if (FindAttribute(BuildName) != NULL)
		return FALSE;

	struct ATTRIBUTE *LastNode;

	LastNode = LastInList;

	if (LastNode == NULL)
	{
		LastNode = new ATTRIBUTE;
		AttribTable = LastNode;
		AttribTable->prev = NULL;
	}
	else
	{
		LastNode->next = new ATTRIBUTE;
		LastNode->next->prev = LastNode;
		LastNode = LastNode->next;
	}
	LastNode->Name = BuildName;
	LastNode->Type = Type;
	LastNode->next = NULL;
	LastNode->Value = CString("");
	LastInList = LastNode;
	TotalAttributes++;

	return TRUE;
}

struct ATTRIBUTE * CAttributeTable::FindAttribute(CString Name)
{
	// Find the named attribute with the 'highest' Level indicator
	CString BuildName;

	struct ATTRIBUTE *SearchNode;
	

	for (SearchNode = LastInList; SearchNode != NULL;)
	{
		if (SearchNode->Name.Left(SearchNode->Name.GetLength()-4) == Name)
			return SearchNode;
		SearchNode = SearchNode->prev;
	}

	return NULL;

}

bool CAttributeTable::RemoveAttribute(CString Name)
{
	struct ATTRIBUTE *DelNode;

	if ((DelNode = FindAttribute(Name)) == NULL)
		return FALSE;

	DeleteAttribute(DelNode);
	return TRUE;
}

void CAttributeTable::DeleteAttribute(struct ATTRIBUTE *Where)
{
	struct ATTRIBUTE *DelNode = Where;

	if (DelNode->next != NULL)
		DelNode->next->prev = DelNode->prev;
	else
		LastInList = DelNode->prev;

	if (DelNode->prev != NULL)
		DelNode->prev->next = DelNode->next;
	else
		AttribTable = DelNode->next;
	TotalAttributes --;

	delete DelNode;
}


bool CAttributeTable::RemoveLevel(int Level)
{
	// Remove any and all attributes whose level is that of the incoming level
	struct ATTRIBUTE *DelNode;
	struct ATTRIBUTE *NextNode;
	CString Search;

	Search.Format("%03d", Level);

	DelNode = AttribTable;
	while (DelNode != NULL)
	{
		if (DelNode->Name.Right(3) == Search)
		{
			NextNode = DelNode->next;
			DeleteAttribute(DelNode);
			DelNode = NextNode;
		}
		else
			DelNode = DelNode->next;
	}
	return TRUE;
}
	
bool CAttributeTable::SetAttrib(CString Name, double Value)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return FALSE;

	// Do type-checking to ensure that Value is in a floatlike format
	AttribNode->Value.Format("%f", Value);
	return TRUE;
}

bool CAttributeTable::SetAttrib(CString Name, bool Value)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return FALSE;

	// Do type-checking to ensure that Value is in a floatlike format
	AttribNode->Value.Format("%s", Value == TRUE ? CString("T") : CString("F"));
	return TRUE;
}

bool CAttributeTable::SetAttrib(CString Name, CString Value)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return FALSE;

	AttribNode->Value = Value;
	return TRUE;
}

CString CAttributeTable::GetAttrib(CString Name)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return CString("");  // Error condition -- throw an exception

	return AttribNode->Value;
}

RETURNTYPE CAttributeTable::GetAttribType(CString Name)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return rtERROR;

	return AttribNode->Type;
}

double  CAttributeTable::GetAttribF(CString Name)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return -99999999.999999999;  // Error condition throw an exception?

	if (AttribNode->Type != rtFLOAT)
		return -99999999.999999999;  // Error condition throw an exception?

	return atof(AttribNode->Value);
}

bool    CAttributeTable::GetAttribB(CString Name)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return FALSE;  // Error condition throw an exception?

	if (AttribNode->Type != rtBOOL)
		return FALSE;  // Error condition throw an exception?

	return (AttribNode->Value == CString("T") ? TRUE : FALSE);
}

CString CAttributeTable::GetAttribS(CString Name)
{
	struct ATTRIBUTE *AttribNode;

	if ((AttribNode = FindAttribute(Name)) == NULL)
		return CString("");  // Error condition throw an exception?

	if (AttribNode->Type != rtSTRING)
		return CString("");  // Error condition throw an exception?

	return AttribNode->Value;
}

int CAttributeTable::GetAttributeCount()
{
	struct ATTRIBUTE *SearchNode;
	int Total = 0;	

	for (SearchNode = AttribTable; SearchNode != NULL;)
	{
		Total++;
		SearchNode = SearchNode->next;
	}

	return Total;
}
	
CString CAttributeTable::GetAttribNameByNumber(int Which)
{
	struct ATTRIBUTE *SearchNode;
	int Total = 0;	

	for (SearchNode = AttribTable; SearchNode != NULL;)
	{
		if (Total == Which)
			break;
		Total++;
		SearchNode = SearchNode->next;
	}
	if (SearchNode == NULL)
		return CString("Error");

	return SearchNode->Name.Left(SearchNode->Name.GetLength()-4);
}


CFunctionTable::CFunctionTable()
{
	Functions = NULL;
}


CFunctionTable::~CFunctionTable()
{
	ClearAll();
}

void CFunctionTable::ClearAll()
{
	struct FUNCTIONNODE *DelNode;
	struct FUNCTIONNODE *CurrNode;

	for (CurrNode = Functions; CurrNode != NULL;)
	{
		DelNode = CurrNode;
		CurrNode = CurrNode->next;

		if (DelNode->ExecTree != NULL)
			delete DelNode->ExecTree;
		delete DelNode;
	}

	Functions = NULL;
}

bool CFunctionTable::AddFunction(CString Name, enum RETURNTYPE Type)
{
	struct FUNCTIONNODE *CurrNode;
	struct FUNCTIONNODE *NextToLast = NULL;

	for (CurrNode = Functions; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		if (CurrNode->Name == Name)
			return FALSE;
		NextToLast = CurrNode;
	}
	
	if (NextToLast == NULL)
	{
		Functions = new FUNCTIONNODE;
		Functions->next = NULL;
		Functions->Type = Type;
		Functions->Name = Name;
		Functions->ExecTree = NULL;
		Functions->NumParameters = 0;
	}
	else
	{
		NextToLast->next = new FUNCTIONNODE;
		NextToLast = NextToLast->next;
		NextToLast->Name = Name;
		NextToLast->Type = Type;
		NextToLast->ExecTree = NULL;
		NextToLast->NumParameters = 0;
		NextToLast->next = NULL;
	}

	return TRUE;
}

struct FUNCTIONNODE * CFunctionTable::FindFunctionNode(CString Name)
{
	struct FUNCTIONNODE *CurrNode;

	for (CurrNode = Functions; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		if (CurrNode->Name == Name)
			return CurrNode;
	}
	return NULL;
}

bool CFunctionTable::FindFunction(CString Name)
{
	if (FindFunctionNode(Name) == NULL)
		return FALSE;

	return TRUE;
}

int CFunctionTable::GetParameterCount(CString Name)
{
	struct FUNCTIONNODE *CurrNode;
	if ((CurrNode = FindFunctionNode(Name)) == NULL)
		return -1;

	return CurrNode->NumParameters;
}

bool CFunctionTable::Compile(CString Name, CString SourceCode)
{
	struct FUNCTIONNODE *CurrNode;

	if ((CurrNode = FindFunctionNode(Name)) == NULL)
		return FALSE;

	CurrNode->ExecTree = new class CGPProgram();
	if (CurrNode->ExecTree->Compile(SourceCode))
		return TRUE;
	else
	{
		delete CurrNode->ExecTree;
		CurrNode->ExecTree = NULL;
		return FALSE;
	}
}

bool CFunctionTable::AddParameter(CString FunctName, CString ParmName, enum RETURNTYPE Type)
{
	struct FUNCTIONNODE *CurrNode;

	if ((CurrNode = FindFunctionNode(FunctName)) == NULL)
		return FALSE;
	
	int ThisParm;
	for (ThisParm = 0; ThisParm < CurrNode->NumParameters; ThisParm++)
	{
		if (CurrNode->Parameters[ThisParm] == ParmName)
			return FALSE;
	}
	CurrNode->Parameters[CurrNode->NumParameters] = ParmName;
	CurrNode->ParameterTypes[CurrNode->NumParameters++] = Type;
	return TRUE;
}

CString CFunctionTable::GetParmName(CString FunctName, int WhichParm)
{
	struct FUNCTIONNODE *CurrNode;
	if ((CurrNode = FindFunctionNode(FunctName)) == NULL)
		return CString("");

	if (CurrNode->NumParameters < WhichParm)
		return CString("");

	if (WhichParm < 0)
		return CString("");
	
	return CurrNode->Parameters[WhichParm];
}

enum RETURNTYPE CFunctionTable::GetParmType(CString FunctName, int WhichParm)
{
	struct FUNCTIONNODE *CurrNode;
	if ((CurrNode = FindFunctionNode(FunctName)) == NULL)
		return rtERROR;

	if (CurrNode->NumParameters < WhichParm)
		return rtERROR;

	if (WhichParm < 0)
		return rtERROR;
	
	return CurrNode->ParameterTypes[WhichParm];
}

enum RETURNTYPE CFunctionTable::GetType(CString FunctName)
{
	struct FUNCTIONNODE *CurrNode;
	if ((CurrNode = FindFunctionNode(FunctName)) == NULL)
		return rtERROR;

	return CurrNode->Type;
}

CString CFunctionTable::GetFunctionName(int WhichFunct)
{
	int index = 0;
	struct FUNCTIONNODE *CurrNode;

	for (CurrNode = Functions; CurrNode != NULL && index != WhichFunct; CurrNode = CurrNode->next)
	{
		if (CurrNode->ExecTree == NULL)
			return CString("");
		index++;
	}
	if (CurrNode == NULL)
		return CString("");

	return CurrNode->Name;
}

bool CFunctionTable::TableValid()
{
	struct FUNCTIONNODE *CurrNode;

	for (CurrNode = Functions; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		if (CurrNode->ExecTree == NULL)
			return FALSE;
	}
	return TRUE;
}

CGPProgram * CFunctionTable::GetFunctionProgram(CString Name)
{
	struct FUNCTIONNODE *CurrNode;

	if ((CurrNode = FindFunctionNode(Name)) == NULL)
		return FALSE;

	if (CurrNode->ExecTree == NULL)
		return NULL;

	return CurrNode->ExecTree;

}

CString CFunctionTable::GetFunctNameByNumber(int Which)
{
	int Current = 0;
	struct FUNCTIONNODE *CurrNode;

	for (CurrNode = Functions; CurrNode != NULL && Current != Which; CurrNode = CurrNode->next)
		Current++;

	if (CurrNode == NULL)
		return (CString("ERROR"));

	if (Current == Which)
		return CurrNode->Name;

	return CString("ERROR");
}

CFunctionSelector::CFunctionSelector()
{
	Selections = NULL;
	TotalSelections = 0;
}


CFunctionSelector::~CFunctionSelector()
{
	ClearAll();
}

void CFunctionSelector::ClearAll()
{
	struct FUNCSELECTIONNODE *DelNode;
	struct FUNCSELECTIONNODE *CurrNode;

	for (CurrNode = Selections; CurrNode != NULL;)
	{
		DelNode = CurrNode;
		CurrNode = CurrNode->next;

		delete DelNode;
	}

	Selections = NULL;
}

bool CFunctionSelector::AddFunction(CString Name)
{
	struct FUNCSELECTIONNODE *CurrNode;
	struct FUNCSELECTIONNODE *NextToLast = NULL;
	int index;

	for (CurrNode = Selections; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		if (CurrNode->Name == Name)
			return FALSE;
		NextToLast = CurrNode;
	}
	
	if (NextToLast == NULL)
	{
		Selections = new FUNCSELECTIONNODE;
		CurrNode = Selections;
		CurrNode->next = NULL;
		CurrNode->Name = Name;
	}
	else
	{
		NextToLast->next = new FUNCSELECTIONNODE;
		CurrNode = NextToLast->next;
		CurrNode->Name = Name;
		CurrNode->next = NULL;
	}

	CurrNode->UserFunction = TRUE; // Defaults to true
	// Search for incoming name in the built-ins list
	for (index = 0; index < TOTALBUILTINFUNCTIONS; index++)
	{
		if (Name == AllBuiltInStrings[index])
			CurrNode->UserFunction = FALSE;
	}

	TotalSelections++;

	return TRUE;
}

bool CFunctionSelector::IsUserFunction(CString Name)
{
	struct FUNCSELECTIONNODE *CurrNode;
	for (CurrNode = Selections; CurrNode != NULL; CurrNode = CurrNode->next)
	{
		if (CurrNode->Name == Name)
			return CurrNode->UserFunction;
	}
	return FALSE;
}

CString CFunctionSelector::GetRandomFunction()
{
	int Selection, index;
	struct FUNCSELECTIONNODE *CurrNode;

	Selection = (int) (RandomGen.Random() * TotalSelections);
	index = 0;
	for (CurrNode = Selections; CurrNode != NULL && index < Selection; CurrNode = CurrNode->next)
		index++;
	if (CurrNode == NULL)
	{
		AfxMessageBox("Catastrophic error in GetRandomFunction()");
		return CString("Catastrophic error!");
	}
	return CurrNode->Name;
}

