//////////////////////////////////////////////
// GAPS Execution Evaluation: Calculate the Fitness of each individual
// CGAPSExecution implementation file
#include "stdafx.h"
#include "GAPS2.h"
#include "EnvironmentDoc.h"
#include "CGAPSIndividual.h"

#include "CGAPSExecution.h"
#include ".\MTRandom\randomc.h"

#include <afxwin.h>

extern TRandomMersenne RandomGen;

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
////EVALUATION
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////

double CGAPSExecution::EvaluateIndividual(class CGAPSIndividual *OurIndividual)
{
	// Returns the fitness of the individual based on the evaluation of the targets
	// Run it through each and every target

	// First, place Individual's tree in the slot where the base program's RPB is

	struct ArgsAndTargets *ThisArgTargPair;
	CString FormatIt;
	CGPProgram GPLabExecution;
	double CurrFitness = 0;
	double StdFitness;
	double FinalAdjustedFitness;
	int AttrIndex;
	int TimesExecuted = 0;


	// Add Specialized attributes for Arguments
	OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg1Type, CString("Argument0"), 0);
	if (OurEnvironment->m_NumArguments > 1)
		OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg2Type, CString("Argument1"), 0);
	if (OurEnvironment->m_NumArguments > 2)
		OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg3Type, CString("Argument2"), 0);
	if (OurEnvironment->m_NumArguments > 3)
		OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg4Type, CString("Argument3"), 0);
	if (OurEnvironment->m_NumArguments > 4)
		OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg5Type, CString("Argument4"), 0);
	if (OurEnvironment->m_NumArguments > 5)
		OurIndividual->AttributeTable->AddAttribute(OurEnvironment->tbl_OurTargets->Arg6Type, CString("Argument5"), 0);
	
	// Add Specialized attribute "Target"
	OurIndividual->AttributeTable->AddAttribute(rtFLOAT, "Target", 0);
	// Add Specialized attribute "Raw"
	OurIndividual->AttributeTable->AddAttribute(rtFLOAT, "Raw", 0);

	GPLabExecution.ProgramTree = OurEnvironment->t_MainProgramTree->ProgramTree;
	GPLabExecution.SetRPBTree(OurIndividual->Tree);

	for (ThisArgTargPair = OurEnvironment->tbl_OurTargets; ThisArgTargPair != NULL; ThisArgTargPair = ThisArgTargPair->next)
	{
		CString FitnessReturnVal;
		// Reset all attributes not marked Persistent
		for (AttrIndex = 0; AttrIndex < OurEnvironment->m_TotalAttributes; AttrIndex++)
			if (!OurEnvironment->a_Attributes[AttrIndex].IsPersistent)
				OurIndividual->AttributeTable->SetAttrib(OurEnvironment->a_Attributes[AttrIndex].Name, CString("0"));

		OurIndividual->AttributeTable->SetAttrib(CString("Argument0"), ThisArgTargPair->Arg1);
		if (OurEnvironment->m_NumArguments > 1)
			OurIndividual->AttributeTable->SetAttrib(CString("Argument1"), ThisArgTargPair->Arg2);
		if (OurEnvironment->m_NumArguments > 2)
			OurIndividual->AttributeTable->SetAttrib(CString("Argument2"), ThisArgTargPair->Arg3);
		if (OurEnvironment->m_NumArguments > 3)
			OurIndividual->AttributeTable->SetAttrib(CString("Argument3"), ThisArgTargPair->Arg4);
		if (OurEnvironment->m_NumArguments > 4)
			OurIndividual->AttributeTable->SetAttrib(CString("Argument4"), ThisArgTargPair->Arg5);
		if (OurEnvironment->m_NumArguments > 5)
			OurIndividual->AttributeTable->SetAttrib(CString("Argument5"), ThisArgTargPair->Arg6);
	
		GPLabExecution.Execute(OurIndividual->AttributeTable, &OurEnvironment->tbl_OurFunctionTable, OurEnvironment->m_AbsoluteTimeLimit);

		OurIndividual->AttributeTable->SetAttrib(CString("Target"), ThisArgTargPair->TargetVal);
		FitnessReturnVal = OurEnvironment->t_RawFitnessTree->Execute(OurIndividual->AttributeTable, &OurEnvironment->tbl_OurFunctionTable, 0);
		CurrFitness += OurEnvironment->t_RawFitnessTree->GetNumExprVal(FitnessReturnVal);
		TimesExecuted++;
	}
		
	GPLabExecution.ProgramTree = NULL;

	FormatIt.Format("%f", CurrFitness);
	OurIndividual->AttributeTable->SetAttrib(CString("Raw"), FormatIt);

	StdFitness =  atof(OurEnvironment->t_StdFitnessTree->Execute(OurIndividual->AttributeTable, &OurEnvironment->tbl_OurFunctionTable, 0));
	StdFitness = fabs(StdFitness);

	// Remove the 'specialized' attributes
	OurIndividual->AttributeTable->RemoveAttribute(CString("Argument0"));
	if (OurEnvironment->m_NumArguments > 1)
		OurIndividual->AttributeTable->RemoveAttribute(CString("Argument1"));
	if (OurEnvironment->m_NumArguments > 2)
		OurIndividual->AttributeTable->RemoveAttribute(CString("Argument2"));
	if (OurEnvironment->m_NumArguments > 3)
		OurIndividual->AttributeTable->RemoveAttribute(CString("Argument3"));
	if (OurEnvironment->m_NumArguments > 4)
		OurIndividual->AttributeTable->RemoveAttribute(CString("Argument4"));
	if (OurEnvironment->m_NumArguments > 5)
		OurIndividual->AttributeTable->RemoveAttribute(CString("Argument5"));

	OurIndividual->AttributeTable->RemoveAttribute(CString("Raw"));
	OurIndividual->AttributeTable->RemoveAttribute(CString("Target"));


	// Calculate and return Adjusted Fitness
	FinalAdjustedFitness = 1.0 / (1.0 + StdFitness);

	// Reset attributes, since they're no longer needed.
	for (AttrIndex = 0; AttrIndex < OurEnvironment->m_TotalAttributes; AttrIndex++)
		OurIndividual->AttributeTable->SetAttrib(OurEnvironment->a_Attributes[AttrIndex].Name, CString("0"));


	// If Shrinkage Mode, Fitness is a calculation of 80% of evaluation plus 20% of the 
	// range between maximum and minimum depths
	if (CurrentExecMode == xm_SHRINKAGE)
	{
		FinalAdjustedFitness = 0.8 * FinalAdjustedFitness;

		FinalAdjustedFitness += 0.2 * 
			((((double) OurEnvironment->m_MaximumDepth - (double) OurIndividual->Tree->GetDepth()) + 1.0 ) /
			(((double) OurEnvironment->m_MaximumDepth - (double) OurEnvironment->m_MinimumDepth) + 1.0));
	}

	return FinalAdjustedFitness;
}



//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
////SORTING ROUTINES 
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
void CGAPSExecution::SortByFitness(CDC *dc)
{
	long SortIndex = OurEnvironment->m_Population;

	// Make Populace a 'Heap' ready for HeapSort
	while (SortIndex > 0)
	{
		Heapify(SortIndex);
		SortIndex--;
		// Keep Populace a Heap-savvy entity
	}
	
	// HeapSort into SortedPopulace;
	for (SortIndex = 1; SortIndex <= OurEnvironment->m_Population; SortIndex++)
	{
		DrawProgressBar(dc, SortIndex/2, OurEnvironment->m_Population);
		CreateIndividualCopy(&SortedPopulace[SortIndex-1], &Populace[0]);

		SwapIndividuals(&Populace[0], &Populace[OurEnvironment->m_Population - SortIndex]);
		HeapifyFromRemove(1L, OurEnvironment->m_Population - SortIndex );
	}

	for (SortIndex = 0; SortIndex < OurEnvironment->m_Population; SortIndex++)
	{
		DrawProgressBar(dc, SortIndex/2 + (OurEnvironment->m_Population/2), OurEnvironment->m_Population);
		if (Populace[SortIndex].Tree != NULL)
			delete Populace[SortIndex].Tree;
		Populace[SortIndex].Tree = NULL;
		Populace[SortIndex].ID = -1;
		Populace[SortIndex].AttributeTable->ClearAll();
	}

	// SortedPopulace now has the population sorted by fitness, ready to do a new population
	// back into Populace

}

void CGAPSExecution::Heapify(long Branch)
{

	long Parent;

	Parent = Branch / 2;
	if (Parent == 0)
		return;

	if (Populace[Parent-1].m_Fitness < Populace[Branch-1].m_Fitness)
	{
		SwapIndividuals(&Populace[Parent-1], &Populace[Branch-1]);
	}

	Heapify(Branch / 2);
}

void CGAPSExecution::HeapifyFromRemove(long Branch, int PopSize)
{
	long LeftChild;
	long RightChild;
	long SwapIndex;
//	CGAPSIndividual TempInd;


	LeftChild = Branch * 2;
	RightChild = Branch * 2 + 1;
	if (LeftChild > PopSize)
		return;

	if (RightChild > PopSize)
		// Swap index is the left one always
		SwapIndex = LeftChild;
	// Otherwise, compete for 'swapness'
	else
	{
		if (Populace[RightChild-1].m_Fitness > Populace[LeftChild-1].m_Fitness)
			SwapIndex = RightChild;
		else
			SwapIndex = LeftChild;
	}

	// Remembering that the index is zero-based and a heap branch node 'ID' is one-based
	if (Populace[Branch-1].m_Fitness > Populace[SwapIndex-1].m_Fitness)
		// All is well; can stop
		return;

	// Perform Swap and heapify again with that child.
	SwapIndividuals(&Populace[SwapIndex-1], &Populace[Branch-1]);

	HeapifyFromRemove(SwapIndex, PopSize);
}

void CGAPSExecution::CreateIndividualCopy(CGAPSIndividual * Destination, CGAPSIndividual * Source)
{
	
	int OldAttrCount;
	int attrindex;
	CString GetAttrVal;

	Destination->ID = Source->ID;
	Destination->m_Fitness = Source->m_Fitness;

	if (Destination->Tree != NULL)
		delete Destination->Tree;
	Destination->Tree = NULL;

	if (Source->Tree != NULL)
		Destination->Tree = Source->Tree->CreateCopy();
	else
		Destination->Tree = NULL;

	Destination->AttributeTable->ClearAll();
	OldAttrCount = Source->AttributeTable->GetAttributeCount();
	for (attrindex = 0; attrindex < OldAttrCount ; attrindex++)
	{
		GetAttrVal = Source->AttributeTable->GetAttribNameByNumber(attrindex);
		Destination->AttributeTable->AddAttribute(Source->AttributeTable->GetAttribType(GetAttrVal), GetAttrVal, 0);
	}
}

void CGAPSExecution::SwapIndividuals(CGAPSIndividual * One, CGAPSIndividual * Two)
{
	CGAPSIndividual TempInd;

	delete TempInd.AttributeTable;

	TempInd.ID             = One->ID;
	TempInd.m_Fitness      = One->m_Fitness;
	TempInd.Tree           = One->Tree;
	TempInd.AttributeTable = One->AttributeTable;

	One->ID                = Two->ID;
	One->m_Fitness         = Two->m_Fitness;
    One->Tree              = Two->Tree;
	One->AttributeTable    = Two->AttributeTable;

	Two->ID                = TempInd.ID;
	Two->m_Fitness         = TempInd.m_Fitness;
	Two->Tree              = TempInd.Tree;
	Two->AttributeTable    = TempInd.AttributeTable;

	TempInd.Tree           = NULL;
	TempInd.AttributeTable = NULL;
}