//////////////////////////////////////////////
// GAPS Execution: Where the magic begins
// CGAPSExecution implementation file
#include "stdafx.h"
#include "GAPS2.h"
#include "EnvironmentDoc.h"
#include "CGAPSIndividual.h"

#include "CGAPSExecution.h"
#include "GAPSHRButton.h"
#include ".\MTRandom\randomc.h"

#include <afxwin.h>

#define SPEEDUP 1

extern TRandomMersenne RandomGen;

CGAPSExecution::CGAPSExecution(CEnvironmentDoc *EnvironmentIn,
							   CWnd *OutputScr) : CObject()
{
	int index;
	if (OutputScr != NULL)
		OurOutputScr = OutputScr;

	OurEnvironment = EnvironmentIn;
	
	Populace       = new CGAPSIndividual [OurEnvironment->m_Population];
	SortedPopulace = new CGAPSIndividual [OurEnvironment->m_Population];
	for (index = 0; index < OurEnvironment->m_Population; index++)
	{
		Populace[index].Tree = NULL;
		SortedPopulace[index].Tree = NULL;
	}

	HonorRoll = new CGAPSIndividual [OurEnvironment->m_HonorRollSize];

	for (index = 0; index < OurEnvironment->m_HonorRollSize; index++)
	{
		HonorRoll[index].Tree = NULL;
		HonorRoll[index].m_Fitness = 0.0;
	}

	HonorRollEntries = 0;
	TotalIndividuals = 0;

	Stopped = FALSE;
}

CGAPSExecution::~CGAPSExecution()
{
	delete [] Populace;
	delete [] SortedPopulace;
	delete [] HonorRoll;
}


bool CGAPSExecution::Launch()
{
	bool Success;

	CClientDC ScreenDC(OurOutputScr);  // Output Window
	CDC OurDC;

	// Create Brushes
	CBrush Background(RGB(0, 0, 0));
	CBrush ProgressBrush(RGB(10, 10, 160));
	CBrush StatBrush(RGB(70, 180, 70));
	CBrush ScaleBrush(RGB(0, 0, 150));
	CBrush BestOfGenBrush(RGB(245, 245, 0));

	// Create Rectangles
	CRect WindowRect;
	CRect ParentRect;
	CRect PopStatRect;
	CRect BestFitRect;
	CRect FitHistRect;
	CRect FitHistHeadRect;
	CRect TimeRect;
	CRect ScaleRect;
	CRect HonorRollRect;

	CString FormatIt;
	
	OurOutputScr->GetClientRect(&WindowRect);
	OurOutputScr->GetParent()->GetClientRect(&ParentRect);
	OurDC.CreateCompatibleDC(&ScreenDC);
	CBitmap ScreenMem;
	ScreenMem.CreateBitmap(WindowRect.Width(), WindowRect.Height(), 1, 16, NULL);
	OurDC.SelectObject(&ScreenMem);

	int CurrentGeneration = 0; 
	double BestFitness = 0.0;
	double ThisFitness = 0.0;
	int CurrentUnproductive = 0;
	int CurrentShrinkage = 0;
	double OldBestFitness = 0.0;
	double BestOfGenFitness;
	long TotalDepth, MinDepth, MaxDepth, GetDepth;
	long TotalNodes, MinNodes, MaxNodes, GetNodes;

	double HistoFitnesses[SPEEDUP];
	int CurrentHisto = 0;

	COleDateTime      StartTime;
	COleDateTime      GenStartTime;
	COleDateTimeSpan  ThisGenTime;
	COleDateTimeSpan  ExecTime;

	// Create Fonts

	CFont BestFitFont;
	BestFitFont.CreateFont(28,0, 0, 0, FW_BOLD, 0, 0, 0,
		DEFAULT_CHARSET, OUT_RASTER_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, 
		DEFAULT_PITCH, "Courier");

	CFont OutputFont;
	OutputFont.CreateFont(13, 0, 0, 0, FW_NORMAL, 0, 0, 0,
		DEFAULT_CHARSET, OUT_CHARACTER_PRECIS, CLIP_DEFAULT_PRECIS, DRAFT_QUALITY, 
		FF_DONTCARE, "Microsoft Sans Serif");

	CFont StatusFont;
	StatusFont.CreateFont(14, 0, 0, 0, FW_BOLD, 0, 0, 0,
		DEFAULT_CHARSET, OUT_CHARACTER_PRECIS, CLIP_DEFAULT_PRECIS, DRAFT_QUALITY, 
		FF_DONTCARE, "Microsoft Sans Serif");

	CFont BestOfGenFont;
	BestOfGenFont.CreateFont(14, 0, 0, 0, FW_NORMAL, 0, 0, 0, 
		DEFAULT_CHARSET, OUT_RASTER_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, 
		FF_DONTCARE, "Times New Roman");


	CDC SaveScreen;
	SaveScreen.CreateCompatibleDC(&OurDC);

	ScreenDC.FillRect(WindowRect, &Background);
	
	PopStatRect.left   = WindowRect.left;
	PopStatRect.top    = WindowRect.top + 30;
	PopStatRect.right  = (WindowRect.right - WindowRect.left) / 2 + WindowRect.left;
	PopStatRect.bottom = PopStatRect.top + 90;
	CRect ExecStatRect;
	ExecStatRect = PopStatRect;
	ExecStatRect.left  = PopStatRect.right;
	ExecStatRect.right = WindowRect.right - 1; 

	ScreenDC.FrameRect(PopStatRect, &StatBrush);
	ScreenDC.FrameRect(ExecStatRect, &StatBrush);

	ScreenDC.SelectObject(&StatBrush);
	ScreenDC.Rectangle(CRect(PopStatRect.left, PopStatRect.top, PopStatRect.right, PopStatRect.top+15));
	ScreenDC.Rectangle(CRect(ExecStatRect.left, ExecStatRect.top, ExecStatRect.right, ExecStatRect.top+15));

	ScreenDC.SelectObject(&StatusFont);
	ScreenDC.SetTextColor(RGB(0, 40, 0));
	ScreenDC.SetBkColor(RGB(30, 180, 30));
	ScreenDC.DrawText("Previous Generation Statistics", CRect(PopStatRect.left, PopStatRect.top+1, PopStatRect.right, PopStatRect.top+16), DT_LEFT);
	ScreenDC.DrawText("Execution Statistics", CRect(ExecStatRect.left, ExecStatRect.top+1, ExecStatRect.right, ExecStatRect.top+16), DT_LEFT);
	BestFitRect.TopLeft().x = ExecStatRect.right - 103;
	BestFitRect.TopLeft().y = ExecStatRect.bottom - 28;
	BestFitRect.BottomRight().x = BestFitRect.left + 100;
	BestFitRect.BottomRight().y = BestFitRect.top + 25;
	ScreenDC.FrameRect(CRect(BestFitRect.left-1, BestFitRect.top-1, BestFitRect.right+1, BestFitRect.bottom+1), &StatBrush);
	ScreenDC.SelectObject(&OutputFont);
	ScreenDC.SetTextColor(RGB(40, 220, 220));
	ScreenDC.SetBkColor(RGB(0, 0, 0));
	ScreenDC.TextOut(ExecStatRect.left + 100, ExecStatRect.bottom - 15, "Current Best Fitness:");

	FitHistRect.top = PopStatRect.bottom+1;
	FitHistRect.left = WindowRect.left + 1;
	FitHistRect.bottom = FitHistRect.top + 120;
	FitHistRect.right = WindowRect.right - 160;

	FitHistHeadRect = FitHistRect;
	FitHistHeadRect.bottom = FitHistHeadRect.top+16;

	ScreenDC.SelectObject(&StatusFont);
	ScreenDC.SetTextColor(RGB(0, 40, 0));
	ScreenDC.SetBkColor(RGB(30, 180, 30));
	ScreenDC.FrameRect(FitHistRect, &StatBrush);
	ScreenDC.FillRect(FitHistHeadRect, &StatBrush);
	ScreenDC.DrawText("Population Per-Individual Fitness", CRect(FitHistHeadRect.left, FitHistHeadRect.top+1, FitHistHeadRect.right, FitHistHeadRect.top+16), DT_LEFT);

	TimeRect.top = FitHistHeadRect.top;
	TimeRect.bottom = FitHistRect.bottom;
	TimeRect.left = FitHistRect.right;
	TimeRect.right = WindowRect.right - 1;

	ScreenDC.FrameRect(TimeRect, &StatBrush);
	ScreenDC.FillRect(CRect(TimeRect.left, TimeRect.top, TimeRect.right, TimeRect.top+16), &StatBrush);
	ScreenDC.DrawText("Exec Times", CRect(TimeRect.left, TimeRect.top, TimeRect.right, TimeRect.top + 16), DT_LEFT);
	

	FitHistRect.top += 16;
	FitHistRect.right -= 70;
	FitHistRect.left += 15;

	ScaleRect.left = FitHistRect.left - 15;
	ScaleRect.top = FitHistRect.top+3; 
	ScaleRect.right = FitHistRect.left -1 ;
	ScaleRect.bottom = FitHistRect.bottom - 1;
	ScreenDC.FillRect(CRect(ScaleRect.left+1, ScaleRect.top, ScaleRect.left + 2, ScaleRect.bottom), &ScaleBrush);
	ScreenDC.FillRect(CRect(ScaleRect.left + 1, ScaleRect.top, ScaleRect.right, ScaleRect.top + 1), &ScaleBrush);
	ScreenDC.FillRect(CRect(ScaleRect.left + 1, ScaleRect.top + 49, ScaleRect.right, ScaleRect.top + 50), &ScaleBrush);
	ScreenDC.FillRect(CRect(ScaleRect.left + 1, ScaleRect.top + 99, ScaleRect.right, ScaleRect.top + 100), &ScaleBrush);

	ScreenDC.FillRect(CRect(FitHistRect.right+19, ScaleRect.top, FitHistRect.right+20, ScaleRect.bottom), &ScaleBrush);
	ScreenDC.FillRect(CRect(FitHistRect.right+1, ScaleRect.top, FitHistRect.right+20, ScaleRect.top + 1), &ScaleBrush);
	ScreenDC.FillRect(CRect(FitHistRect.right+1, ScaleRect.top + 49, FitHistRect.right+20, ScaleRect.top + 50), &ScaleBrush);
	ScreenDC.FillRect(CRect(FitHistRect.right+1, ScaleRect.top + 99, FitHistRect.right+20, ScaleRect.top + 100), &ScaleBrush);

	HonorRollRect.top = FitHistRect.bottom + 1;
	HonorRollRect.left = WindowRect.left + 1;
	HonorRollRect.right = WindowRect.right - 1;
	HonorRollRect.bottom = WindowRect.bottom - 1;
	ScreenDC.FrameRect(HonorRollRect, &StatBrush);
	ScreenDC.FillRect(CRect(HonorRollRect.left, HonorRollRect.top+1, HonorRollRect.right, HonorRollRect.top+16), &StatBrush);
	ScreenDC.DrawText("Honor Roll", CRect(HonorRollRect.left, HonorRollRect.top+1, HonorRollRect.right, HonorRollRect.top+16), DT_LEFT);

	// Times of Execution
	StartTime = COleDateTime::GetCurrentTime();

	// Copy current bitmap into the Outputwindow for keeping
	SaveScreen.SelectObject(CBitmap::FromHandle(((CStatic *) OurOutputScr)->GetBitmap()));
	SaveScreen.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);

	// Initial Population
	DrawHeaderOut(&ScreenDC, CString("INITIAL POPULATION GENERATION"));

	// Initiate Execution
	// Generation of Initial Population
	ScreenDC.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);

	if (Genesis(&ScreenDC))
	{ 
		CurrentExecMode = xm_NORMAL;

		ScreenDC.SelectObject(&OutputFont);
		ScreenDC.SetBkColor(0x00000000);
		FormatIt.Format("Start Time: %02d:%02d:%02d", StartTime.GetHour(), StartTime.GetMinute(), StartTime.GetSecond());
		ScreenDC.DrawText(FormatIt, CRect(TimeRect.left+5, TimeRect.top+18, TimeRect.right-1, TimeRect.top+32), DT_LEFT);

		///////////////////////////////////////////////////////////////
		// Genetic Execution
		while (CheckAfxMessages() && !Stopped)
		{
			CString FormatIt;

			CurrentGeneration++;

			ScreenDC.SelectObject(&OutputFont);
			ScreenDC.SetBkColor(0x00000000);
			ScreenDC.SetTextColor(RGB(40, 220, 220));
			FormatIt.Format("Current Generation:%d", CurrentGeneration);
			GenStartTime = COleDateTime::GetCurrentTime();
			ScreenDC.DrawText(FormatIt, CRect(ExecStatRect.left + 3, ExecStatRect.top + 18, ExecStatRect.right - 3, ExecStatRect.top + 32), DT_LEFT);

			ScreenDC.FillRect(CRect(FitHistRect.left + 1, FitHistRect.top + 1, FitHistRect.right - 1, FitHistRect.bottom -1), &Background);


			TotalAdjPopFitness = 0.0;
			BestOfGenFitness = 0.0;
			TotalDepth = 0;
			MinDepth = 9999;
			MaxDepth = 0;
			TotalNodes = 0;
			MinNodes = 9999;
			MaxNodes = 0;

			// Do Fitness
			FormatIt.Format("FITNESS TESTING%s", CurrentExecMode == xm_SHRINKAGE ? " - SHRINKAGE MODE" : "");
			DrawHeaderOut(&ScreenDC, FormatIt);

			long TestIndividual;

			OldBestFitness = BestFitness;

			// Fitness Testing of entire population
			for (TestIndividual = 0; 
					TestIndividual < OurEnvironment->m_Population && CheckAfxMessages() &&!Stopped; 
					TestIndividual ++)
			{

				if (CurrentHisto == SPEEDUP || TestIndividual == OurEnvironment->m_Population)
				{
					DrawProgressBar(&ScreenDC, TestIndividual, OurEnvironment->m_Population);
					ScreenDC.SelectObject(&OutputFont);
					ScreenDC.SetTextColor(RGB(40, 195, 220));
					ExecTime = COleDateTime::GetCurrentTime() - StartTime;
					FormatIt.Format("Total Exec Time: %02d:%02d:%02d", ExecTime.GetHours(), ExecTime.GetMinutes(), ExecTime.GetSeconds());
					ScreenDC.DrawText(FormatIt, CRect(TimeRect.left+5, TimeRect.top + 34, TimeRect.right-1, TimeRect.top+48), DT_LEFT);
				}

				ThisFitness = EvaluateIndividual(&Populace[TestIndividual]);
				Populace[TestIndividual].m_Fitness = ThisFitness;
				TotalAdjPopFitness += ThisFitness;
				GetDepth = Populace[TestIndividual].Tree->GetDepth(0);
				TotalDepth += GetDepth;
				if (GetDepth < MinDepth)
					MinDepth = GetDepth;
				if (GetDepth > MaxDepth)
					MaxDepth = GetDepth;
				GetNodes = Populace[TestIndividual].Tree->GetNodeCount();
				TotalNodes += GetNodes;
				if (GetNodes < MinNodes)
					MinNodes = GetNodes;
				if (GetNodes > MaxNodes)
					MaxNodes = GetNodes;


				if (ThisFitness > BestOfGenFitness)
				{
					ScreenDC.SelectObject(&BestOfGenFont);
					ScreenDC.SetTextColor(RGB(255, 255, 0));
					ScreenDC.SetBkColor (RGB(0, 0, 0));
					BestOfGenFitness = ThisFitness;
					ScreenDC.FillRect(CRect(FitHistRect.right+21, FitHistRect.top, FitHistHeadRect.right-1, FitHistRect.bottom-1), &Background);
					ScreenDC.FillRect(CRect(FitHistRect.right+21, FitHistRect.bottom - (int) (100 * BestOfGenFitness)-2, FitHistRect.right + 26,FitHistRect.bottom - (int) (100 * BestOfGenFitness) - 1), &BestOfGenBrush); 
					FormatIt.Format("%1.4f", BestOfGenFitness);
					int Top;
					ScreenDC.DrawText(FormatIt, CRect(FitHistRect.right + 26, 
												   Top = (FitHistRect.bottom - (int) (100 * BestOfGenFitness) - 12) > FitHistRect.top + 1 ? (FitHistRect.bottom - (int) (100 * BestOfGenFitness) - 12) : FitHistRect.top + 1, 
												   FitHistHeadRect.right, 
												   Top + 11), DT_LEFT);
				}


				// Do Histogram Output of Fitness -- every ten individuals for speed
				if (CurrentHisto == SPEEDUP || TestIndividual == OurEnvironment->m_Population)
				{
					int CHIndex;
					for (CHIndex = 0; CHIndex < CurrentHisto; CHIndex++)
					{
						if (TestIndividual - SPEEDUP + CHIndex > (FitHistRect.right - FitHistRect.left))
						{
							// 'scroll' rect left one		
							ScreenDC.ScrollDC(-1, 0, &FitHistRect, &FitHistRect, NULL, NULL);
							ScreenDC.FillRect(CRect(FitHistRect.right-1, FitHistRect.top, FitHistRect.right+1, FitHistRect.bottom -1), &Background);
							ScreenDC.SetPixel(FitHistRect.right - SPEEDUP - 2 + CHIndex, FitHistRect.bottom - (int) (100 * HistoFitnesses[CHIndex]) - 2, RGB(255, 60 + (int) (200 * HistoFitnesses[CHIndex]), 20));
						}
						else
							ScreenDC.SetPixel(FitHistRect.left/* - SPEEDUP*/ - 2 + CHIndex + TestIndividual, FitHistRect.bottom - (int) (100 * HistoFitnesses[CHIndex]) - 2, RGB(255, 60+ (int) (175 * HistoFitnesses[CHIndex]), 20));
					}

					CurrentHisto = 0;
				}
				else
				{
					HistoFitnesses[CurrentHisto++] = ThisFitness;
				}

				if (ThisFitness >= BestFitness)
				{
					BestFitness = ThisFitness;
					if (fabs(BestFitness - 1.000) < OurEnvironment->m_FitnessThreshold)
					{
						ScreenDC.SetTextColor(RGB(150, 255, 40));
					}
					else
						ScreenDC.SetTextColor(RGB(40, 195, 220));

					FormatIt.Format("%1.4f", BestFitness);
					ScreenDC.SelectObject(&BestFitFont);
					ScreenDC.SetBkColor(RGB(0,0,0));
					ScreenDC.DrawText(FormatIt, BestFitRect, DT_CENTER);
				}
				
				// Do an insertion on the Honor Roll, if fitness warrants it.
				int hindex;
				if (ThisFitness > HonorRoll[OurEnvironment->m_HonorRollSize - 1].m_Fitness ||
					(CurrentExecMode == xm_SHRINKAGE && 
					 ThisFitness == HonorRoll[OurEnvironment->m_HonorRollSize - 1].m_Fitness &&
					 Populace[TestIndividual].Tree->GetDepth() < HonorRoll[OurEnvironment->m_HonorRollSize - 1].Tree->GetDepth()))
				{
					for (hindex = OurEnvironment->m_HonorRollSize - 1; hindex > 0; hindex--)
					{
						CreateIndividualCopy(&HonorRoll[hindex], &HonorRoll[hindex-1]);
						if (ThisFitness < HonorRoll[hindex-1].m_Fitness)
							break;
					}
					CreateIndividualCopy(&HonorRoll[hindex], &Populace[TestIndividual]);
				} 
				
				if (CurrentHisto == SPEEDUP || TestIndividual == OurEnvironment->m_Population)
					ScreenDC.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);
			}

			DrawProgressBar(&ScreenDC, 0, 1);

			
			if (CurrentExecMode == xm_SHRINKAGE)
				CurrentShrinkage ++;
				
			if (OldBestFitness == BestFitness)
				CurrentUnproductive++;	
			else
				CurrentUnproductive = 0;



			ScreenDC.SelectObject(&OutputFont);
			ScreenDC.SetBkColor(0x00000000);
			ScreenDC.SetTextColor(RGB(40, 220, 220));
			

			ThisGenTime = COleDateTime::GetCurrentTime() - GenStartTime;
			FormatIt.Format("Generation Exec Time: %02d:%02d:%02d", ThisGenTime.GetHours(), ThisGenTime.GetMinutes(), ThisGenTime.GetSeconds());
			ScreenDC.DrawText(FormatIt, CRect(TimeRect.left+5, TimeRect.top + 50, TimeRect.right-1, TimeRect.top+64), DT_LEFT);
			
			// Stop button hit
			if (Stopped)
				break;

			FormatIt.Format("Current Unproductive Gens: %d of %d       ", CurrentUnproductive, OurEnvironment->m_MaxUnproductiveGens);
			ScreenDC.DrawText(FormatIt, CRect(ExecStatRect.left + 3, ExecStatRect.top + 32, ExecStatRect.right - 3, ExecStatRect.top + 47), DT_LEFT);

			if (CurrentExecMode == xm_SHRINKAGE)
			{
				FormatIt.Format("Current Shrinkage Gens: %d of %d       ", CurrentShrinkage, OurEnvironment->m_ShrinkageGenerations);
				ScreenDC.DrawText(FormatIt, CRect(ExecStatRect.left + 3, ExecStatRect.top + 46, ExecStatRect.right - 3, ExecStatRect.top + 61), DT_LEFT);
			}

			FormatIt.Format("Average Individual Fitness: %1.4f", (double) TotalAdjPopFitness / (OurEnvironment->m_Population == 0 ? 1 : OurEnvironment->m_Population));
			ScreenDC.DrawText(FormatIt, CRect(PopStatRect.left + 3, PopStatRect.top + 18, PopStatRect.right - 3, PopStatRect.top + 32), DT_LEFT);

			FormatIt.Format("Average Individual Node Depth: %2.2f (Min %d, Max %d)    ", (double) TotalDepth / (OurEnvironment->m_Population == 0 ? 1 : OurEnvironment->m_Population), MinDepth, MaxDepth);
			ScreenDC.DrawText(FormatIt, CRect(PopStatRect.left + 3, PopStatRect.top + 32, PopStatRect.right - 3, PopStatRect.top + 47), DT_LEFT);

			FormatIt.Format("Average Invidual Node Size: %2.2f (Min %d, Max %d)    ", (double) TotalNodes / (OurEnvironment->m_Population == 0 ? 1 : OurEnvironment->m_Population), MinNodes, MaxNodes);
			ScreenDC.DrawText(FormatIt, CRect(PopStatRect.left + 3, PopStatRect.top + 46, PopStatRect.right - 3, PopStatRect.top + 61), DT_LEFT);


			///////////////////////////////////////////////////////////////////
			//**TERMINATION CRITERIA**/////////////////////////////////////////
			///////////////////////////////////////////////////////////////////
			if (fabs(BestFitness - 1.000) < OurEnvironment->m_FitnessThreshold ||
				CurrentUnproductive == OurEnvironment->m_MaxUnproductiveGens ||
				CurrentShrinkage == OurEnvironment->m_ShrinkageGenerations)
			{
				// If best honor roll candidate has depth of environment's minimum,
				// it's as good as it can be.
				if (HonorRoll[0].Tree->GetDepth() == OurEnvironment->m_MinimumDepth)
					break;

				// If not already shrinkage mode, switch to shrinkage mode
				if (CurrentExecMode != xm_SHRINKAGE)
				{
					CurrentExecMode = xm_SHRINKAGE;

					// Increase mutation pressure
					OurEnvironment->m_MutationRate = OurEnvironment->m_MutationRate * 5.0;

					// Reset Unproductive Generations
					CurrentUnproductive = 0;

					CurrentShrinkage =0;
				}
				else
				{
					if (CurrentUnproductive == OurEnvironment->m_MaxUnproductiveGens ||
						CurrentShrinkage == OurEnvironment->m_ShrinkageGenerations)
					{
						// It's done it's best.  What we got is what we got.
						break;
					}
				}
			}
			///////////////////////////////////////////////////////////////////
			//**END TERMINATION CRITERIA**/////////////////////////////////////
			///////////////////////////////////////////////////////////////////


			// Sort population based on fitness -- Goes into SortedPopulace
			ScreenDC.SelectObject(&OutputFont);
			DrawHeaderOut(&ScreenDC, CString("SORTING"));
			SortByFitness(&ScreenDC);

			// Do Genetic Functions, upon the newly sorted SortedPopulace
			DrawHeaderOut(&ScreenDC, CString("BREEDING NEW GENERATION"));
			ScreenDC.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);

			// Newest Generation is created into Populace
			BreedNewGeneration(&ScreenDC);
			DrawProgressBar(&ScreenDC, 0, 1);
			ScreenDC.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);

		}	
	}

	if (Stopped)
	{
		DrawHeaderOut(&ScreenDC, CString("**SIMULATION CANCELED**"));
		Success = FALSE;
	}
	else
	{
		Success = TRUE;
		FormatIt.Format("SIMULATION COMPLETE: %d TOTAL GENERATIONS", CurrentGeneration);
		DrawHeaderOut(&ScreenDC, FormatIt);
	}
	
	ScreenDC.BitBlt(0, 0, WindowRect.Width(), FitHistRect.bottom+1, &ScreenDC, 0, 0, SRCCOPY);

	// Copy current bitmap into the Outputwindow for keeping
	SaveScreen.SelectObject(CBitmap::FromHandle(((CStatic *) OurOutputScr)->GetBitmap()));
	SaveScreen.BitBlt(0, 0, WindowRect.Width(), WindowRect.Height(), &ScreenDC, 0, 0, SRCCOPY);

	return Success;
	
}

void CGAPSExecution::DrawHeaderOut(CDC *dc, CString &HeaderStr)
{
	CRect WindowRect;
	CRect ModeRect;
	CBrush ModeBrush(RGB(0, 80, 0));
	CFont ModeFont;
	ModeFont.CreateFont(14, 12, 0, 0, FW_BOLD, 0, 0, 0, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, 
		FF_ROMAN, "Arial");

	OurOutputScr->GetClientRect(&WindowRect);

	ModeRect.TopLeft().x = WindowRect.left;
	ModeRect.TopLeft().y = WindowRect.top ;
	ModeRect.BottomRight().x = WindowRect.right;
	ModeRect.BottomRight().y = ModeRect.top + 15;

	dc->FillRect(ModeRect, &ModeBrush);
	dc->SelectObject(&ModeFont);
	dc->SetTextColor(RGB(40, 220, 255));
	dc->SetBkColor(RGB(0, 80, 0));

	dc->DrawText(HeaderStr, ModeRect, DT_CENTER);
}

void CGAPSExecution::DrawProgressBar(CDC *OurDC, long Current, long OutOf)
{
	CString FormatIt;
	CBrush ProgressBrush(RGB(10, 10, 160));
	CBrush EraseBrush(RGB(0, 0, 0));

	CRect ProgressRect;
	CRect WindowRect;

	int TextLeft;

	OurOutputScr->GetClientRect(&WindowRect);

	CFont StatusFont;
	StatusFont.CreateFont(14, 0, 0, 0, FW_BOLD, 0, 0, 0,
		DEFAULT_CHARSET, OUT_CHARACTER_PRECIS, CLIP_DEFAULT_PRECIS, DRAFT_QUALITY, 
		FF_DONTCARE, "Microsoft Sans Serif");


	ProgressRect.TopLeft().x = WindowRect.left;
	ProgressRect.TopLeft().y = WindowRect.top + 15 ;
	ProgressRect.BottomRight().x = WindowRect.left;
	ProgressRect.BottomRight().y = ProgressRect.top + 15;

	double PercentComplete;
	PercentComplete =  ((double) Current / (double) OutOf);
	ProgressRect.right = (int) (PercentComplete * WindowRect.right);
	TextLeft = ProgressRect.right + 1;
	if (TextLeft >= WindowRect.right - 37)
		TextLeft = WindowRect.right-38;

	OurDC->FillRect(ProgressRect, &ProgressBrush);
	OurDC->FillRect(CRect(ProgressRect.right, ProgressRect.top, WindowRect.right, ProgressRect.bottom), &EraseBrush);
	OurDC->SetBkColor(RGB(0, 0, 0));
	OurDC->SetTextColor(RGB(100, 120, 255));
	OurDC->SelectObject(&StatusFont);
	if (Current != 0L)
	{
		FormatIt.Format("%2.0f%c", PercentComplete * 100, '%');
		OurDC->TextOut(TextLeft, ProgressRect.top+2, FormatIt);
	}	
}