////////////////////////////////////////////////////////
// GAPS Execution: Where the magic happens
// Header file

class CGAPSExecution : public CObject
{
public:
	CGAPSExecution (CEnvironmentDoc * OurEnvironment, 
		            CWnd * OutputScr);
	~CGAPSExecution();

	bool Launch();

	bool Stopped;
	enum EXECUTIONMODE CurrentExecMode;

	// Initial Generations
	bool Genesis(CClientDC *dc);
	void CreateAttributeInfo(class CGAPSIndividual * Individual);

	class CSubTree * GenerateTreeNode(int CurrDepth, enum RETURNTYPE SubTreeType);
	class CSubTree * GenerateTreeGrow(int CurrDepth, enum RETURNTYPE SubTreeType);
	class CSubTree * GenerateTreeFull(int CurrDepth, enum RETURNTYPE SubTreeType);

	class CSubTree * GenerateRandomFunction(int CurrDepth, enum RETURNTYPE SubTreeType);
	class CSubTree * GenerateRandomTerminal(enum RETURNTYPE SubTreeType);

	class CSubTree * BuildBuiltIn(CString Name, int CurrDepth, enum RETURNTYPE SubTreeType);

	// Individual Helper Functions
	void CreateIndividualCopy(CGAPSIndividual * Destination, CGAPSIndividual * Source);
	void SwapIndividuals(CGAPSIndividual * One, CGAPSIndividual * Two);

	// Fitness Evaluation
	double EvaluateIndividual(class CGAPSIndividual *OurIndividual);
	void SortByFitness(CDC *dc);
	void Heapify(long Branch);
	void HeapifyFromRemove(long Branch, int PopSize);

	// Breed Next Generation
	long SelectRandomIndividual();
	void BreedNewGeneration(CClientDC *dc);
	void CrossOver(CGAPSIndividual * IndividualA, CGAPSIndividual * IndividualB);
	void Mutation (CGAPSIndividual * Individual);   // 
	
	// Check validity of crossed-over trees
	bool IsValidIndividual(CGAPSIndividual * Individual);
	bool IsNodeValid(CSubTree * Node, enum RETURNTYPE ExpectedType, class CAttributeTable * IndAttribs);
	enum RETURNTYPE GetArgumentType(CString InValue);

	// Screen Drawing Helper Funcs
	void DrawHeaderOut(CDC *dc, CString &HeaderStr);
	void DrawProgressBar(CDC *OurDC, long Current, long OutOf);

	class CGAPSIndividual * HonorRoll;
	

private:
	CWnd * OurOutputScr;
	CEnvironmentDoc * OurEnvironment;

	class CGAPSIndividual * Populace;
	class CGAPSIndividual * SortedPopulace;

	long TotalIndividuals;
	int HonorRollEntries;
	double TotalAdjPopFitness;  // Total Adjusted Fitness of all individuals

};