//////////////////////////////////////////////
// GAPS Genetic Manipulation: Where the system 
// genetically manipulates the trees
// CGAPSExecution implementation file
#include "stdafx.h"
#include "GAPS2.h"
#include "EnvironmentDoc.h"
#include "CExecEnv.h"
#include "CGAPSIndividual.h"

#include "CGAPSExecution.h"
#include ".\MTRandom\randomc.h"

#include <afxwin.h>

extern TRandomMersenne RandomGen;

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
////NEW GENERATION BREEDING
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
void CGAPSExecution::BreedNewGeneration(CClientDC *dc)
{
	// Using TotalAdjPopFitness and the Selection Strategy, create a whole new population
	// based upon our sorted population
	// 
	// The current Cross-Over vs. straight reproduction is 90-10, but will be a 
	// user-modifiable parameter after getting it going.

	long NewPopIndex;
	long SelectedIndividual1;
	long SelectedIndividual2;
	CString FormatIt;
	bool KeepTrying;
	
	for (NewPopIndex = 0; NewPopIndex < OurEnvironment->m_Population && CheckAfxMessages(); NewPopIndex++)
	{
		DrawProgressBar(dc, NewPopIndex, OurEnvironment->m_Population);

			if (RandomGen.Random() < 0.9) // Crossover of two individuals
			{
				// Create individuals
				CGAPSIndividual NewIndivid1;
				CGAPSIndividual NewIndivid2;

				CreateAttributeInfo(&NewIndivid1);
				CreateAttributeInfo(&NewIndivid2);

				NewIndivid1.ID = TotalIndividuals++;
				NewIndivid2.ID = TotalIndividuals++;

				NewIndivid1.AttributeTable->ResetValues();
				NewIndivid2.AttributeTable->ResetValues();

				// Perform Cross-over of the two individuals, two 'new' individuals
				// Copy new individuals from Selected Individuals
				int Tries = 0;
				KeepTrying = TRUE;
				do 
				{
			
					if (NewIndivid1.Tree != NULL)
						delete NewIndivid1.Tree;
					NewIndivid1.Tree = NULL;
					if (NewIndivid2.Tree != NULL)
						delete NewIndivid2.Tree;
					NewIndivid2.Tree = NULL;


					SelectedIndividual1 = SelectRandomIndividual();
					SelectedIndividual2 = SelectRandomIndividual();

					NewIndivid1.Tree = SortedPopulace[SelectedIndividual1].Tree->CreateCopy();
					NewIndivid2.Tree = SortedPopulace[SelectedIndividual2].Tree->CreateCopy();

					// Then perform operation "CrossOver"
					CrossOver(&NewIndivid1, &NewIndivid2);


					// Now, perhaps perform mutation on Individual 1
					if (RandomGen.Random() < OurEnvironment->m_MutationRate / 100)
					{
						Mutation(&NewIndivid1);
					}

					// Mutation on Individual 2
					if (RandomGen.Random() < OurEnvironment->m_MutationRate / 100)
					{
						Mutation(&NewIndivid2);
					}

					KeepTrying = !IsValidIndividual(&NewIndivid1);
					KeepTrying = KeepTrying || !IsValidIndividual(&NewIndivid2);

					Tries++;
					
				} while ( KeepTrying && Tries < 500);

				if (Tries == 500)
				{
					// Couldn't create a good new entity, so we just copy the old one in and mutate it
					CreateIndividualCopy(&Populace[NewPopIndex], &SortedPopulace[SelectedIndividual1]);
					Mutation(&Populace[NewPopIndex]);
				}
				else
				{
					CreateIndividualCopy(&Populace[NewPopIndex], &NewIndivid1);
					if (NewPopIndex < OurEnvironment->m_Population - 1)  // If population has room
						CreateIndividualCopy(&Populace[++NewPopIndex], &NewIndivid2);
				}
			}
			else  // Straight reproduction
			{

				SelectedIndividual1 = SelectRandomIndividual();
				CreateIndividualCopy(&Populace[NewPopIndex], &SortedPopulace[SelectedIndividual1]);
				if (RandomGen.Random() < OurEnvironment->m_MutationRate / 100)
				{
					Mutation(&Populace[NewPopIndex]);
				}
			}

	}

	DrawHeaderOut(dc, CString("CLEANING OLD POPULATION"));
	// Kill Old Sorted Populace
	for (NewPopIndex = 0; NewPopIndex < OurEnvironment->m_Population ; NewPopIndex++)
	{
		delete SortedPopulace[NewPopIndex].Tree;
		SortedPopulace[NewPopIndex].Tree = NULL;
		SortedPopulace[NewPopIndex].ID = -1;
		SortedPopulace[NewPopIndex].AttributeTable->ClearAll();
	}

  
}


//////////////////////////////////////////////////////////////////////////////////
////SELECTION PROCESS
//////////////////////////////////////////////////////////////////////////////////
// Determines an individual to be selected for genetic manipulation
// based on Selection Strategy
long CGAPSExecution::SelectRandomIndividual()
{
	double FitnessSummation;
	double RouletteWheel;
	long Selected;
	long Candidate;
	double SelectedFitness;

	long Group1Total;
	long Group2Total;
	int index;

	switch (OurEnvironment->m_SelectionStrategy)
	{
	case ss_FITNESS:
		// Straight 'Roulette-wheel' Fitness-Proportionate Selection 
		RouletteWheel = RandomGen.Random() * TotalAdjPopFitness;
		FitnessSummation = 0.0;
		for (Selected = 0L; Selected < OurEnvironment->m_Population; Selected++)
		{
			FitnessSummation += SortedPopulace[Selected].m_Fitness;
			if (FitnessSummation > RouletteWheel)
				break;
		}
		return Selected;  

	case ss_TOURN:
		// 'Tournament-style' selection

		SelectedFitness = 0.0;
		for (index = 0; index < OurEnvironment->m_TournamentSize; index++)
		{
			Candidate = (long) (RandomGen.Random() * OurEnvironment->m_Population);

			if (SortedPopulace[Candidate].m_Fitness > SelectedFitness)
			{
				Selected = Candidate;
				SelectedFitness = SortedPopulace[Candidate].m_Fitness;
			}
		}

		return Selected;

	case ss_OVERSEL:
		// 'Greedy' Overselection... Select from Group 1 80% of the time. (Group 1 is the top X% of 
		// population in Fitness; Group 2 is the remaining)
		Group1Total = (long) (OurEnvironment->m_Population * (OurEnvironment->m_SelectionParmNum / 100));
		Group2Total = OurEnvironment->m_Population - Group1Total;

		if ((RandomGen.Random() * 10) < 8.0)
			// Select from Group 1
			return (long) (RandomGen.Random() * Group1Total);
		else
			return (long) ((long) (RandomGen.Random() * Group2Total) + Group1Total);
		break;

	}

	return 0L;
}

//////////////////////////////////////////////////////////////////////////////////
////CROSSOVER
//////////////////////////////////////////////////////////////////////////////////
void CGAPSExecution::CrossOver(CGAPSIndividual * IndividualA, CGAPSIndividual * IndividualB)
{
	// Do actual crossover 

	// Select a random point in each individual's tree and swap them
	// To do this, select the pointer to a random child from a random point in 
	// the tree of the first individual and make it point to a random node in the
	// tree of the second invidiual;
	// the point where the second's child was pointing to is changed to point to
	// where the first was instead.

	// If either of the new trees violates the depth maximum, that tree is discarded
	// and the corresponding individual will not have its tree modified.

	// 3/23/2004 Updated to make Crossover semi-contextually aware (e.g., if a target root of the 
	// selected subtree requires a boolean value, then the crossed-over subtree will also be a boolean-
	// returning tree)

	CSubTree * NewTreeA = NULL;
	CSubTree * NewTreeB = NULL;

	CSubTree * ParentNodeA;
	CSubTree * ParentNodeB;

	CSubTree * NewSubTreeA;
	CSubTree * NewSubTreeB;

	int ChildA;
	int ChildB;

	int SearchDepth;
	int DepthChoice;

	NewTreeA = IndividualA->Tree->CreateCopy();
	NewTreeB = IndividualB->Tree->CreateCopy();

	enum RETURNTYPE NodeAContext;
	enum RETURNTYPE NodeBContext;

	// Find a random parent node and a random child node for each
	ParentNodeA = NewTreeA;
	SearchDepth = 1;
	ChildA = 0;
//	DepthChoice = (int) (RandomGen.Random() * (OurEnvironment->m_MaximumDepth-2)) +1;
	DepthChoice = (int) (RandomGen.Random() * NewTreeA->GetDepth()) + 1;
	while (ParentNodeA->GetChildCount() > 0)
	{
		// Don't alter the children of MAPs.
		ChildA = (int) (RandomGen.Random() * ParentNodeA->GetChildCount());
		NodeAContext = OurEnvironment->t_MainProgramTree->GetChildContext(ParentNodeA->GetChild(ChildA)->GetType(), 
			ParentNodeA->GetChild(ChildA)->GetVal(),
			ChildA,
			&OurEnvironment->tbl_OurFunctionTable);
		if (ParentNodeA->GetChild(ChildA)->GetChildCount() == 0)
			break;
		if (SearchDepth == DepthChoice)
			break;
		SearchDepth++;
		ParentNodeA = ParentNodeA->GetChild(ChildA);
	}

	if (ParentNodeA->GetChild(ChildA)->GetType() == ttUSRATTRIB)
		NodeAContext = IndividualA->AttributeTable->GetAttribType(ParentNodeA->GetChild(ChildA)->GetVal());
	if (NodeAContext == rtERROR)
	{
		if (ParentNodeA->GetChild(ChildA)->GetVal().Left(8) == CString("Argument"))
		{
			NodeAContext = GetArgumentType(ParentNodeA->GetChild(ChildA)->GetVal());
		}
	}
	
	ParentNodeB = NewTreeB;
	SearchDepth = 1;
//	DepthChoice = (int) (RandomGen.Random() * (OurEnvironment->m_MaximumDepth-2)) +1;
	ChildB = 0;
	DepthChoice = (int) (RandomGen.Random() * NewTreeB->GetDepth()) + 1;
	while (ParentNodeB->GetChildCount() > 0)
	{
		ChildB = (int) (RandomGen.Random() * ParentNodeB->GetChildCount());
		NodeBContext = OurEnvironment->t_MainProgramTree->GetChildContext(ParentNodeB->GetChild(ChildB)->GetType(), 
			ParentNodeB->GetChild(ChildB)->GetVal(),
			ChildB,
			&OurEnvironment->tbl_OurFunctionTable);
		if (NodeBContext == NodeAContext)
		{
			if (SearchDepth >= DepthChoice)
				break;
		}
		if (ParentNodeB->GetChild(ChildB)->GetChildCount() == 0)
			break;
		SearchDepth++;
		ParentNodeB = ParentNodeB->GetChild(ChildB);
	}

	if (ParentNodeB->GetChild(ChildB)->GetType() == ttUSRATTRIB)
		NodeBContext = IndividualB->AttributeTable->GetAttribType(ParentNodeB->GetChild(ChildB)->GetVal());
	if (NodeBContext == rtERROR)
	{
		if (ParentNodeB->GetChild(ChildB)->GetVal().Left(8) == CString("Argument"))
		{
			NodeBContext = GetArgumentType(ParentNodeB->GetChild(ChildB)->GetVal());
		}
	}
	
	NewSubTreeA = ParentNodeB->GetChild(ChildB)->CreateCopy();
	NewSubTreeB = ParentNodeA->GetChild(ChildA)->CreateCopy();

	delete ParentNodeA->GetChild(ChildA);
	ParentNodeA->SetChild(ChildA, NewSubTreeA);

	if (NodeBContext == NodeAContext)
	{
		delete ParentNodeB->GetChild(ChildB);
		ParentNodeB->SetChild(ChildB, NewSubTreeB);
	}
	else
	{
		if (NewSubTreeB != NULL)
			delete NewSubTreeB;
		NewSubTreeB = NULL;
	}
	

	if (NewTreeA->GetDepth() <= OurEnvironment->m_MaximumDepth &&
		NewTreeA->GetDepth() >= OurEnvironment->m_MinimumDepth)
	{
		delete IndividualA->Tree;
		IndividualA->Tree = NewTreeA;
	}
	else
		delete NewTreeA;

	if (NewTreeB->GetDepth() <= OurEnvironment->m_MaximumDepth &&
		NewTreeB->GetDepth() >= OurEnvironment->m_MinimumDepth)
	{
		delete IndividualB->Tree;
		IndividualB->Tree = NewTreeB;
	}
	else
		delete NewTreeB;

}



//////////////////////////////////////////////////////////////////////////////////
////MUTATION
//////////////////////////////////////////////////////////////////////////////////
void CGAPSExecution::Mutation(CGAPSIndividual * Individual)
{
	CSubTree * NewTree = NULL;
	CSubTree * ParentNode;
	int Child;

	int SearchDepth;
	int DepthChoice;

	enum RETURNTYPE NodeContext;


	// 'Snip' a random node in the tree, replacing it with a new growth
	NewTree = Individual->Tree->CreateCopy();

	ParentNode = NewTree;
	SearchDepth = 1;
	DepthChoice = (int) (RandomGen.Random() * OurEnvironment->m_MaximumDepth);
	while (TRUE)
	{
		Child = (int) (RandomGen.Random() * ParentNode->GetChildCount());
		NodeContext = OurEnvironment->t_MainProgramTree->GetChildContext(ParentNode->GetChild(Child)->GetType(), 
			ParentNode->GetChild(Child)->GetVal(),
			Child,
			&OurEnvironment->tbl_OurFunctionTable);
		if (ParentNode->GetChild(Child)->GetChildCount() == 0)
			break;
		if (SearchDepth >= DepthChoice)
			break;
		SearchDepth++;
		ParentNode = ParentNode->GetChild(Child);
	}

	if (ParentNode->GetChild(Child)->GetType() == ttUSRATTRIB)
		NodeContext = Individual->AttributeTable->GetAttribType(ParentNode->GetChild(Child)->GetVal());

	if (NodeContext == rtERROR)
	{
		if (ParentNode->GetChild(Child)->GetVal().Left(8) == CString("Argument"))
		{
			NodeContext = GetArgumentType(ParentNode->GetChild(Child)->GetVal());
		}
	}
	
	delete ParentNode->GetChild(Child);
	
	ParentNode->SetChild(Child, GenerateTreeNode(SearchDepth+1, NodeContext));

	delete Individual->Tree;
	Individual->Tree = NewTree;
}

//////////////////////////////////////////////////////////////////////////////////
////Validity Check 
//////////////////////////////////////////////////////////////////////////////////
bool CGAPSExecution::IsValidIndividual(CGAPSIndividual * Individual)
{
	// Ensure that a crossover-created tree is valid.
	return IsNodeValid(Individual->Tree, OurEnvironment->m_RPBContext, Individual->AttributeTable);
}

bool CGAPSExecution::IsNodeValid(CSubTree * Node, enum RETURNTYPE ExpectedType, class CAttributeTable *IndAttribs)
{

	if (Node == NULL)
		return FALSE;

	switch (Node->GetType())
	{
	case ttFLOAT:
		return (ExpectedType == rtFLOAT ||
			    ExpectedType == rtEXPRESSION ||
				ExpectedType == rtBOOL);
	case ttBOOL:
		return (ExpectedType == rtBOOL);

	case ttSTRING:
		return (ExpectedType == rtSTRING);

	case ttADD:
	case ttSUBTRACT:
	case ttMULTIPLY:
	case ttDIVIDE:
	case ttMOD:
		if (ExpectedType != rtFLOAT &&
			ExpectedType != rtEXPRESSION)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
		if (!IsNodeValid(Node->GetChild(0), rtEXPRESSION, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtEXPRESSION, IndAttribs))
			return FALSE;
		return TRUE;

	case ttNEGATIVE:
		if (ExpectedType != rtFLOAT)
			return FALSE;
		if (Node->GetChild(0) == NULL)
			return FALSE;
		return (IsNodeValid(Node->GetChild(0), rtFLOAT, IndAttribs));

	case ttNOT:
		if (ExpectedType != rtBOOL &&
			ExpectedType != rtEXPRESSION)
			return FALSE;
		if (Node->GetChild(0) == NULL)
			return FALSE;
		return (IsNodeValid(Node->GetChild(0), rtBOOL, IndAttribs));

	case ttAND:
	case ttOR:
		if (ExpectedType != rtBOOL &&
			ExpectedType != rtEXPRESSION)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
		return (IsNodeValid(Node->GetChild(0), rtBOOL, IndAttribs));


	case ttCOMPAREOP:
		if (ExpectedType != rtBOOL &&
			ExpectedType != rtEXPRESSION)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(0), rtBOOL, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtBOOL, IndAttribs))
			return FALSE;
		return TRUE;

	case ttNUMFUNCT:
		if (ExpectedType != rtFLOAT &&
			ExpectedType != rtEXPRESSION)
			return FALSE;
		if (Node->GetChild(0) == NULL)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(0), rtEXPRESSION, IndAttribs))
			return FALSE;
		return TRUE;		

	case ttUSRFUNCTION:
		// check Type of function
		enum RETURNTYPE ReadType;
		int parmindex;
		ReadType = OurEnvironment->tbl_OurFunctionTable.GetType(Node->GetVal());
		switch (ReadType)
		{
		case rtFLOAT:
			if (ExpectedType != rtFLOAT && ExpectedType != rtEXPRESSION && ExpectedType != rtBOOL)
				return FALSE;
			break;
		case rtBOOL:
			if (ExpectedType != rtBOOL)
				return FALSE;
			break;
		case rtSTRING:
			if (ExpectedType != rtSTRING)
				return FALSE;
			break;
		case rtVOID:
			if (ExpectedType != rtVOID)
				return FALSE;
			break;
		}
		// Check all parameters
		for (parmindex = 0; parmindex < OurEnvironment->tbl_OurFunctionTable.GetParameterCount(Node->GetVal()); parmindex++)
		{
			ReadType = OurEnvironment->tbl_OurFunctionTable.GetParmType(Node->GetVal(), parmindex);
			if (Node->GetChild(parmindex) == NULL)
				return FALSE;
			if (!IsNodeValid(Node->GetChild(parmindex), ReadType, IndAttribs))
				return FALSE;
		}
		return TRUE;

	case ttUSRATTRIB:
		// Check Type of attribute
		int index;
		for (index = 0; index < IndAttribs->GetAttributeCount(); index++)
		{
			if (IndAttribs->GetAttribNameByNumber(index) == Node->GetVal())
				break;
		}

		if (Node->GetVal().Left(8) == CString("Argument"))
		{
			switch (GetArgumentType(Node->GetVal()))
			{
			case rtFLOAT:
				return (ExpectedType == rtFLOAT ||
						ExpectedType == rtEXPRESSION ||
						ExpectedType == rtBOOL);
			case rtBOOL:
				return (ExpectedType == rtBOOL);

			case rtSTRING:
				return (ExpectedType == rtSTRING);
			}
		}

		if (index == IndAttribs->GetAttributeCount())
			return FALSE;
		switch (IndAttribs->GetAttribType(Node->GetVal()))
		{
		case rtFLOAT:
			return (ExpectedType == rtFLOAT ||
				    ExpectedType == rtEXPRESSION ||
					ExpectedType == rtBOOL);
		case rtBOOL:
			return (ExpectedType == rtBOOL);

		case rtSTRING:
			return (ExpectedType == rtSTRING);
		}
		return FALSE;

	// Each of the statements
	case ttFNNOOP:
		return (ExpectedType == rtVOID);

	case ttFNIF:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL ||
			Node->GetChild(2) == NULL)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(0), rtBOOL, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtVOID, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 3)
			if (!IsNodeValid(Node->GetChild(3), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;
	
	case ttFNWHILE:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(0), rtBOOL, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtVOID, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 2)
			if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;

	case ttFNDOUNTIL:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(0), rtVOID, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtBOOL, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 2)
			if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;

	case ttFNASSIGN:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (Node->GetChild(0)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (Node->GetChild(0)->GetVal().Left(8) == CString("Argument"))
			return FALSE;
		// Only a statement will be incorrect
		if (IsNodeValid(Node->GetChild(1), rtVOID, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 2)
			if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;
		
	case ttFNPUSH:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (Node->GetChild(0)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (Node->GetChild(0)->GetVal().Left(8) == CString("Argument"))
			return FALSE;
		// Only a statement will be incorrect
		if (IsNodeValid(Node->GetChild(1), rtVOID, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 2)
			if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;

	case ttFNPOP:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL)
			return FALSE;
		if (Node->GetChild(0)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (Node->GetChild(1)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (Node->GetChild(1)->GetVal().Left(8) == CString("Argument"))
			return FALSE;
		if (Node->GetChildCount() > 2)
			if (!IsNodeValid(Node->GetChild(2), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;

	case ttFNREPLACE:
		if (ExpectedType != rtVOID)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL ||
			Node->GetChild(2) == NULL ||
			Node->GetChild(3) == NULL)
			return FALSE;
		if (Node->GetChild(0)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtSTRING, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(2), rtEXPRESSION, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(3), rtEXPRESSION, IndAttribs))
			return FALSE;
		if (Node->GetChildCount() > 4)
			if (!IsNodeValid(Node->GetChild(4), rtVOID, IndAttribs))
				return FALSE;
		return TRUE;
		
	case ttFNGETAT:
		if (ExpectedType != rtSTRING)
			return FALSE;
		if (Node->GetChild(0) == NULL ||
			Node->GetChild(1) == NULL ||
			Node->GetChild(2) == NULL)
			return FALSE;
		if (Node->GetChild(0)->GetType() != ttUSRATTRIB)
			return FALSE;
		if (!IsNodeValid(Node->GetChild(1), rtEXPRESSION, IndAttribs))
			return FALSE;
		if (!IsNodeValid(Node->GetChild(2), rtEXPRESSION, IndAttribs))
			return FALSE;
		return TRUE;

	}

	return FALSE;
}

enum RETURNTYPE CGAPSExecution::GetArgumentType(CString InValue)
{
	// Find the type of the argument
	if (InValue == CString("Argument0"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;
	if (InValue == CString("Argument1"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;
	if (InValue == CString("Argument2"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;
	if (InValue == CString("Argument3"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;
	if (InValue == CString("Argument4"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;
	if (InValue == CString("Argument5"))
		return OurEnvironment->tbl_OurTargets->Arg1Type;

	return rtERROR;
}