//////////////////////////////////////////////
// GAPS Tree Growth: How a tree is generated
// CGAPSExecution implementation file
#include "stdafx.h"
#include "GAPS2.h"
#include "EnvironmentDoc.h"
#include "CGAPSIndividual.h"

#include "CGAPSExecution.h"
#include ".\MTRandom\randomc.h"

#include <afxwin.h>

extern TRandomMersenne RandomGen;



//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
////Generation -- Growth of Random Trees
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////

bool CGAPSExecution::Genesis(CClientDC * dc)
{
	long index;

	CurrentExecMode = xm_GENESIS;
	for (index = 0; index < OurEnvironment->m_Population && CheckAfxMessages(); index++)
	{	
		DrawProgressBar(dc, index, OurEnvironment->m_Population);

		if (!CheckAfxMessages())
			return FALSE;
		
		switch (OurEnvironment->m_GenerationStrategy)
		{
		case gs_GENGROW:
			Populace[TotalIndividuals].Tree = GenerateTreeGrow(1, OurEnvironment->m_RPBContext);
			break;
		case gs_GENFULL:
			Populace[TotalIndividuals].Tree = GenerateTreeFull(1, OurEnvironment->m_RPBContext);
			break;
		case gs_GENHALF:  // Randomize half-and-half.  Note that the population won't be 
			              // exactly one-half of one and one-half of the other.
			if (RandomGen.Random() < 0.5 ) 
				Populace[TotalIndividuals].Tree = GenerateTreeGrow(1, OurEnvironment->m_RPBContext);
			else
				Populace[TotalIndividuals].Tree = GenerateTreeFull(1, OurEnvironment->m_RPBContext);
			break;
		}

		Populace[TotalIndividuals].ID = TotalIndividuals;
		CreateAttributeInfo(&Populace[TotalIndividuals]);
		TotalIndividuals++;
	
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
////GenerateTreeNode
////This function is used both during Genesis and throughout the remaining
////Mutation-based tree growths.  
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
class CSubTree * CGAPSExecution::GenerateTreeNode(int CurrDepth, enum RETURNTYPE SubTreeType)
{
	// Creation of 'Full' trees only happens during the initial Genesis of the initial
	// Population.
	if (CurrentExecMode != xm_GENESIS)
		return GenerateTreeGrow(CurrDepth, SubTreeType);

	if (OurEnvironment->m_GenerationStrategy == gs_GENGROW || 
		(OurEnvironment->m_GenerationStrategy == gs_GENHALF && RandomGen.Random() > 0.5))
	{
		return GenerateTreeGrow(CurrDepth, SubTreeType);
	}
	else
	{
		return GenerateTreeFull(CurrDepth, SubTreeType);
	}
}

class CSubTree * CGAPSExecution::GenerateTreeGrow(int CurrDepth, enum RETURNTYPE SubTreeType)
{
	// If the current tree's depth has not yet reached minimum, it will automatically
	// be a function.
	if (CurrDepth < OurEnvironment->m_MinimumDepth)
	{
		return GenerateRandomFunction(CurrDepth, SubTreeType);
	}

	// If a current tree's depth has reached its maximum, it will automatically be
	// a terminal, or if it's a VOID subtype requirement, it will return a NULL.
	if (CurrDepth == OurEnvironment->m_MaximumDepth)
	{
		return GenerateRandomTerminal(SubTreeType);
	}

	// Depending on Growth Mode, Select a random determination of functions or terminals
	// During 'normal' and 'Genesis' mode, the system will select Functions 33% of the time.  During
	// 'shrinkage' mode, the system will reduce that selection to 20%.
	if (RandomGen.Random() > (CurrentExecMode == xm_SHRINKAGE ? 0.80 : 0.667))
	{
		return GenerateRandomFunction(CurrDepth, SubTreeType);
	}
	else 
	{
		return GenerateRandomTerminal(SubTreeType);
	}
}

class CSubTree * CGAPSExecution::GenerateTreeFull(int CurrDepth, enum RETURNTYPE SubTreeType)
{
	// If a current tree's depth has not reached its maximum, it must be a function 
	if (CurrDepth != OurEnvironment->m_MaximumDepth)
		return GenerateRandomFunction(CurrDepth, SubTreeType);
	return GenerateRandomTerminal(SubTreeType);
}


class CSubTree * CGAPSExecution::GenerateRandomFunction(int CurrDepth, enum RETURNTYPE SubTreeType)
{
	// Could be a user-entered function, or it could be a selected built-in function
	CString SelectedFunction;
	CSubTree * Node;
	int index;
	enum RETURNTYPE NewContextType;

	bool KeepTrying = TRUE;
	int FailsafeAttempts = 0;

	while (KeepTrying) 
	{
		enum RETURNTYPE ThisFunctType;
		SelectedFunction = OurEnvironment->tbl_OurFunctionSelections.GetRandomFunction();
		if (OurEnvironment->tbl_OurFunctionSelections.IsUserFunction(SelectedFunction))
			ThisFunctType = OurEnvironment->tbl_OurFunctionTable.GetType(SelectedFunction);
		else
			ThisFunctType = OurEnvironment->t_RPBNode.GetBuiltinContext(SelectedFunction);

		if (ThisFunctType == SubTreeType)
		{
			KeepTrying = FALSE;
			break;
		}
		if (SubTreeType == rtEXPRESSION &&
			(ThisFunctType == rtFLOAT || ThisFunctType == rtBOOL))
		{
			KeepTrying = FALSE;
			break;
		}
		FailsafeAttempts++;
		if (FailsafeAttempts > 255)  // Breakpoint
		{
			// Could not create a function, so create a base terminal in the context
			Node = new CSubTree; 
			if (SubTreeType == rtFLOAT || SubTreeType == rtEXPRESSION)
				Node->SetType(ttFLOAT);
			if (SubTreeType == rtBOOL)
				Node->SetType(ttBOOL);
			if (SubTreeType == rtSTRING)
				Node->SetType(ttSTRING);
			Node->SetVal(CString("1"));
			return Node;
		}
	}
	
	if (OurEnvironment->tbl_OurFunctionSelections.IsUserFunction(SelectedFunction))
	{ //////////// User-Entered Function
		Node = new CSubTree;
		Node->SetType(ttUSRFUNCTION);
		Node->SetVal(SelectedFunction);
		for (index = 0; index < OurEnvironment->tbl_OurFunctionTable.GetParameterCount(SelectedFunction); index++)
		{
			// Note, we use the GetChildContext method of t_RPBNode simply because it's 
			// there and we don't have to instantiate a new CGPProgram.
			NewContextType = OurEnvironment->t_RPBNode.GetChildContext(ttUSRFUNCTION, SelectedFunction, index, &OurEnvironment->tbl_OurFunctionTable);
			Node->SetChild(index, GenerateTreeNode(CurrDepth+1, NewContextType)); 
		}

		//////////////////////////////////////////////////////////////////////////
		// Add 'next statement' if the latest node was a void (Statement)
		// Note that any function returned that is in a formula or other calculation will,
		// upon execution, ignore the 'next statement' thread.
		if (OurEnvironment->tbl_OurFunctionTable.GetType(SelectedFunction) == rtVOID &&
			(RandomGen.Random() > 0.33 || CurrDepth < OurEnvironment->m_MinimumDepth))
		{
			Node->SetNextChild(GenerateTreeNode(CurrDepth+1, rtVOID));
		}
		return Node;
	}
	else  
	{ //////////// Built-In
		Node = BuildBuiltIn(SelectedFunction, CurrDepth, SubTreeType);
		return Node;
	}

	return NULL;
}

//////////////////////////////////////////////////////////////////
// BuildBuiltIn
// Creates function subtrees for built-in functions selected
class CSubTree * CGAPSExecution::BuildBuiltIn(CString Name, int CurrDepth, enum RETURNTYPE SubTreeType)
{
	CSubTree * Node;
	CSubTree * ChildNode;
	bool Statement;
	int Selected;

	Node = new CSubTree;
	Statement = FALSE;

	Node->SetVal(Name);

	// Go through the cases
	if (Name == CString(":="))
	{
		Node->SetType(ttFNASSIGN);
		ChildNode = new CSubTree;
		ChildNode->SetType(ttUSRATTRIB);
		do {
			Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
			ChildNode->SetVal(OurEnvironment->a_Attributes[Selected].Name);
		} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
		Node->SetChild(0, ChildNode);
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtEXPRESSION));
		Statement = TRUE;
	}

	if (Name == CString("Push"))
	{
		Node->SetType(ttFNPUSH);
		ChildNode = new CSubTree;
		ChildNode->SetType(ttUSRATTRIB);
		do {
			Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
			ChildNode->SetVal(OurEnvironment->a_Attributes[Selected].Name);
		} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
		Node->SetChild(0, ChildNode);
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtEXPRESSION));
		Statement = TRUE;
	}

	if (Name == CString("Pop"))
	{
		Node->SetType(ttFNPOP);
		ChildNode = new CSubTree;
		ChildNode->SetType(ttUSRATTRIB);
		do {
			Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
			ChildNode->SetVal(OurEnvironment->a_Attributes[Selected].Name);
		} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
		Node->SetChild(0, ChildNode);
		ChildNode = new CSubTree;
		ChildNode->SetType(ttUSRATTRIB);
		do {
			Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
			ChildNode->SetVal(OurEnvironment->a_Attributes[Selected].Name);
		} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
		Node->SetChild(1, ChildNode);
		Statement = TRUE;
	}

	if (Name == CString("Replace"))
	{
		Node->SetType(ttFNPUSH);
		ChildNode = new CSubTree;
		ChildNode->SetType(ttUSRATTRIB);
		do {
			Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
			ChildNode->SetVal(OurEnvironment->a_Attributes[Selected].Name);
		} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
		Node->SetChild(0, ChildNode);
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtEXPRESSION));
		Node->SetChild(2, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(3, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Statement = TRUE;
	}

	if (Name == CString("If"))
	{
		Node->SetType(ttFNIF);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtVOID));
		if (Node->GetChild(1) == NULL)
			int DEBUGThis = 1;
		ChildNode = new CSubTree(ttFNNOOP, CString(""));
		Node->SetChild(2, ChildNode);
		Statement = TRUE;
	}

	if (Name == CString("If-Else"))
	{
		Node->SetType(ttFNIF);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtVOID));
		Node->SetChild(2, GenerateTreeNode(CurrDepth+1, rtVOID));
		Statement = TRUE;
	}

	if (Name == CString("While"))
	{
		Node->SetType(ttFNWHILE);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtVOID));
		Statement = TRUE;
	}

	if (Name == CString("DoUntil"))
	{
		Node->SetType(ttFNDOUNTIL);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtVOID));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Statement = TRUE;
	}

	if (Name == CString("sin") || 
		Name == CString("cos") || 
		Name == CString("tan") || 
		Name == CString("log") || 
		Name == CString("sqrt") || 
		Name == CString("round") || 
		Name == CString("trunc") || 
		Name == CString("abs"))
	{
		Node->SetType(ttNUMFUNCT);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("not"))
	{
		Node->SetType(ttNOT);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
	}

	if (Name == CString(">") || 
		Name == CString("<") || 
		Name == CString("=") || 
		Name == CString("!=") || 
		Name == CString(">=") || 
		Name == CString("<="))
	{
		Node->SetType(ttCOMPAREOP);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("+"))
	{
		Node->SetType(ttADD);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("-"))
	{
		Node->SetType(ttSUBTRACT);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("*"))
	{
		Node->SetType(ttMULTIPLY);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("/"))
	{
		Node->SetType(ttDIVIDE);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("%"))
	{
		Node->SetType(ttMOD);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtFLOAT));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtFLOAT));
	}

	if (Name == CString("&&"))
	{
		Node->SetType(ttAND);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtBOOL));
	}

	if (Name == CString("||"))
	{
		Node->SetType(ttOR);
		Node->SetChild(0, GenerateTreeNode(CurrDepth+1, rtBOOL));
		Node->SetChild(1, GenerateTreeNode(CurrDepth+1, rtBOOL));
	}
	
	//////////////////////////////////////////////////////////
	// If random function is a statement, there may be more statements generated.
	if (Statement && (RandomGen.Random() > 0.33 || CurrDepth < OurEnvironment->m_MinimumDepth))
		// Generate Next possible
		Node->SetNextChild(GenerateTreeNode(CurrDepth+1, rtVOID));

	return Node;
}

class CSubTree * CGAPSExecution::GenerateRandomTerminal(enum RETURNTYPE SubTreeType)
{
	int TotalRandomTermPossibles;
	int SelectedType = 0;
	int Selected;
	bool Finished = FALSE;
	CSubTree * Node;
	CString FormatIt;
	bool ZeroParmFunctExists = FALSE;


	int index;
	for (index = 0; index < OurEnvironment->m_TotalUserEnteredFunctions; index++)
	{
		if (OurEnvironment->tbl_OurFunctionTable.GetParameterCount(OurEnvironment->tbl_OurFunctionTable.GetFunctNameByNumber(index)) == 0)
			ZeroParmFunctExists = TRUE;
	}

	if (SubTreeType != rtVOID)
	{
		TotalRandomTermPossibles = 
			OurEnvironment->m_TermsSelectedArgs + 
			OurEnvironment->m_TermsSelectedAttributes + 
			OurEnvironment->m_TermsSelectedBools +
			OurEnvironment->m_TermsSelectedRandNums +
			OurEnvironment->m_TermsSelectedStrings +
			ZeroParmFunctExists;

		if (TotalRandomTermPossibles == 0)
		{
			AfxMessageBox("No selectable terminal, and no zero-parameter function exists to return a terminal.");
			return NULL;
		}


	}

	Node = new CSubTree;

	// In the case of too deep a tree still trying to generate a VOID function or terminal..
	if (SubTreeType == rtVOID)
	{
		Node->SetType(ttFNNOOP);
		Node->SetVal(CString(""));
		return Node;
	}

	while (!Finished)
	{
		SelectedType = (int) (RandomGen.Random() * 6);
		switch (SelectedType)
		{
		case 0:  // Random system input-output Argument
			if (OurEnvironment->m_TermsSelectedArgs)
			{
				Node->SetType(ttUSRATTRIB);
				Selected = (int) (RandomGen.Random() * OurEnvironment->m_NumArguments);
				FormatIt.Format("Argument%d", Selected);
				Node->SetVal(FormatIt);
				Finished = TRUE;
			}
			break;
		case 1:  // Random system Attribute
			if (OurEnvironment->m_TermsSelectedAttributes)
			{
				Node->SetType(ttUSRATTRIB);
				do {
					Selected = (int) (RandomGen.Random() * OurEnvironment->m_TotalAttributes);
					Node->SetVal(OurEnvironment->a_Attributes[Selected].Name);
				} while (!OurEnvironment->a_Attributes[Selected].IsSelectable);
				Finished = TRUE;
			}
			break;
		case 2:  // Random Numeric Constant
			if (OurEnvironment->m_TermsSelectedRandNums)
			{
				double RandomNum;
				Node->SetType(ttFLOAT);
				
				RandomNum = (RandomGen.Random() * 
							((OurEnvironment->m_TermsSelectedNumHigh - OurEnvironment->m_TermsSelectedNumLow) + 1 ))
							+ OurEnvironment->m_TermsSelectedNumLow;
				FormatIt.Format("%f", RandomNum);
				FormatIt = FormatIt.Mid(0, (FormatIt.Find(".") + OurEnvironment->m_TermsSelectedNumDecimals));
				Node->SetVal(FormatIt);
				Finished = TRUE;
			}
			break;
		case 3:  // Random Boolean Constant
			if (OurEnvironment->m_TermsSelectedBools)
			{
				Node->SetType(ttBOOL);
				Node->SetVal((RandomGen.Random() > 0.5) ? CString("TRUE") : CString("FALSE"));
				Finished = TRUE;
			}
			break;

		case 4:  // Random String Constant
			if (OurEnvironment->m_TermsSelectedStrings)
			{
				int Length;
				char RandomChar;
				Node->SetType(ttSTRING);
				Length = (int) (RandomGen.Random() * 
					((OurEnvironment->m_TermsSelectedCharLenHigh - OurEnvironment->m_TermsSelectedCharLenLow) + 1) )
					+ OurEnvironment->m_TermsSelectedCharLenLow;
				FormatIt = CString("");
				for (index = 0; index < Length; index++)
				{	
					RandomChar = (char) (RandomGen.Random() *
						((OurEnvironment->m_TermsSelectedCharHigh - OurEnvironment->m_TermsSelectedCharLow) - 1))
						+ OurEnvironment->m_TermsSelectedCharLenLow;
					FormatIt += RandomChar;
				}
				Node->SetVal(FormatIt);
				Finished = TRUE;
			}
			break;
		case 5:  // Zero-parameter user function (if one exists)
			{
				Node->SetType(ttUSRFUNCTION);
				Node->SetVal(OurEnvironment->tbl_OurFunctionSelections.GetRandomFunction());
				if (OurEnvironment->tbl_OurFunctionTable.FindFunction(Node->GetVal()) &&
					OurEnvironment->tbl_OurFunctionTable.GetType(Node->GetVal()) != rtVOID)
				{
					if (OurEnvironment->tbl_OurFunctionTable.GetParameterCount(Node->GetVal()) == 0)
						Finished = TRUE;
				}
			}
			break;
		}
	} 
	return Node;
}


void CGAPSExecution::CreateAttributeInfo(CGAPSIndividual * Individual)
{
	int AttrIndex;
	CString FormatIt;

	// Create Attribute Table 
	for (AttrIndex = 0; AttrIndex < OurEnvironment->m_TotalAttributes; AttrIndex++)
	{
		enum RETURNTYPE SetType;
		if (OurEnvironment->a_Attributes[AttrIndex].Type == CString("Number"))
			SetType = rtFLOAT;
		if (OurEnvironment->a_Attributes[AttrIndex].Type == CString("Boolean"))
			SetType = rtBOOL;
		if (OurEnvironment->a_Attributes[AttrIndex].Type == CString("String"))
			SetType = rtSTRING;
		Individual->AttributeTable->AddAttribute(SetType, OurEnvironment->a_Attributes[AttrIndex].Name, 0);
	}

}
