//////////////////////////////////////////////////////
// CGPProgram:  GP Lab's Programming Language Class  /
//////////////////////////////////////////////////////
#include "CExecEnv.h"

#ifndef CGPProgramDefined 
#define CGPProgramDefined 1
class CGPProgram  : public CObject 
{

public:

	CGPProgram ();
	CGPProgram (CString &NameIn);
	~CGPProgram();

public:
	// Source Code for an instantiated program/function is stored here.
	CString m_SourceCode; 

	// Human-readable name
	CString m_Name;    

	// The Program's Execution Tree
	class CSubTree * ProgramTree;

	// The Program's Genetically Manipulated Tree
	class CSubTree * RPBTree;
	void SetRPBTree(CSubTree * NewRPBTree);

	void SetReturnType(enum RETURNTYPE InType);
	enum RETURNTYPE GetReturnType();

	bool Compile ();
	bool Compile (CString SourceIn);

	CString Execute(class CAttributeTable * AttribTblIn, class CFunctionTable * FunctionTblIn, long MaxEvaluationsAllowed);

	class CSubTree * FindRPB(class CSubTree * StartNode);
	enum RETURNTYPE GetRPBContext(class CSubTree * StartNode, CFunctionTable * FunctionTblin);
	enum RETURNTYPE GetChildContext(enum TOKENTYPE CallerType, CString CallerName, int WhichChild, CFunctionTable * FunctionTblin);
	enum RETURNTYPE GetBuiltinContext(CString WhichBuiltin);

	bool CheckLegalFunctions(class CSubTree * StartNode, class CFunctionTable * FunctionTblIn);
	bool CheckUsedAttributes(class CSubTree * StartNode, struct UserAttributes AttributeTbl[], int TotalAttribs);

	bool   GetBoolExprVal(CString ReturnedVal);
	double GetNumExprVal(CString ReturnedVal);

	int NumArguments;

protected:

	// Temporary Build Structures are here
	RETURNTYPE ReturnType;

	// Execution-oriented routines here
	CString ExecRecursive(class CSubTree * Node);
	void ExecValues(CString Attrib, class CSubTree *Node);


	void GetNextToken();

	// Compile Methods here
	CSubTree * CompileProgram();
	CSubTree * CompileComplexStatement();
	CSubTree * CompileStatements();
	CSubTree * CompileStatement();
	CSubTree * CompileValueExprs();
	CSubTree * CompileValues();
	CSubTree * CompileAttribute();
	CSubTree * CompileExpression();
	CSubTree * CompileTopLevelExpression();
	CSubTree * CompileFullTerm();
	CSubTree * CompileFullFactor();
	CSubTree * CompileNumExpression();
	CSubTree * CompileNumUnaryFunction();
	CSubTree * CompileUserFunction();
	CSubTree * CompileTerm();
	CSubTree * CompileFactor();
	CSubTree * CompileStringExpr();

	// Statements
	CSubTree * CompileAssign();
	CSubTree * CompileMap();
	CSubTree * CompileIf();
	CSubTree * CompileWhile();
	CSubTree * CompileDoUntil();
	CSubTree * CompilePush();
	CSubTree * CompilePop();
	CSubTree * CompileReplace();
	CSubTree * CompileReturn();
	CSubTree * CompileDeclare();

	// Parse String
	CString ParseString;
	CString CurrentToken;

	// Attribute Table
	class CAttributeTable * AttribTable;
	class CFunctionTable  * FunctionTable;

	int CurrentFunctLevel;   // For determining the scope of local function variables

	long CurrentEvaluations;  // If a program has run out of its maximum evaluations (as
	                         // determined by the Environmental parameter of the same
							 // description), it must halt.
	long MaximumEvaluations;


};


////////////////////////////////////////////////////////////
// GP-Lab Programming Language full Grammar
// <Program>           ::= <ComplexStatement> 
//                       | <Statements>
//                       | <Expression>
// <ComplexStatement>  ::= {<Statements>}
// <Statements>        ::= <Statement> ; <Statements> 
//                       | <ComplexStatement>
//                       | NULL
// <Statement>         ::= <Attribute> := <Expression>
//                       | Map ( <Attribute> ) { <ValueExprs> }
//                       | If ( <BooleanExpr> ) <ComplexStatement>
//                       | If ( <BooleanExpr> ) <ComplexStatement> Else <ComplexStatement>
//                       | While ( <BooleanExpr> ) <ComplexStatement>
//                       | DoUntil( <BooleanExpr> ) <ComplexStatement>
//                       | Push( <Attribute> , <Expression> )
//                       | Pop( <Attribute> , <Attribute>)
//                       | Replace( <Attribute> , <StringExpression> , <NumExpression> , <NumExpression> )
//                       | Declare (<Attribute>, [Number | Boolean | String])
// <ValueExprs>        ::= <Values> <ValueExprs> | NULL
// <Values>            ::= Value <Expression> <ComplexStatement>
// <Attribute>         ::= &<c><C>
// <Expression>        ::= <TopLevelExpression>
//                       | " <StringExpression> "
// <StringExpression>  ::= "<A>"
//                       | <Attribute>
//                       | GetAt ( <Attribute> , <NumExpression> , <NumExpression> )
// <A>                 ::= **(Any string of alphanumeric characters)**
// <C>                 ::= c<C> | NULL
// <c>                 ::= {Any one character from a-z or A-Z}
// <TopLevelExpression>::= not <TopLevelExpression>
//                       | <FullTerm> [|| <TopLevelExpression>]...
// <FullTerm>          ::= <FullFactor> [&& <TopLevelExpression>]...
// <FullFactor>        ::= <NumExpression> [[< | > | != | <> | >= | <= | =] <TopLevelExpression>]...
// <NumExpression>     ::= <Term> [[+ | -] <TopLevelExpression>]...
// <Term>              ::= <Factor> [[* | / | ^ | %] <TopLevelExpression>] ...
// <Factor>            ::= <Attribute>
//                       | <Number>
//                       | - <Number>
//                       | <MathUnaryFunction> ( <NumExpression> )
//                       | <UserFunction> ( [<Attribute> [, <Attribute>]...])
//                       | TRUE | FALSE
//                       | ( <TopLevelExpression> )
// <MathUnaryFunction> ::= sin | cos | tan | log | sqrt | round | abs | trunc
// <UserFunction>      ::= <CharStartedAlphaNumString>
// <CharStartedAlphaNumString> 
//                     ::= <C>[<c> | <d>]...
// <Number>            ::= D.D
//                       | D.DeD
//                       | D.De-D
// <D>                 ::= d<D> | NULL
// d                   ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9

class CompilationException
{
public:
	CompilationException(CString & ErrorMsg);
	~CompilationException();

	CString GetError();

protected:
	CString OurErrorMessage;
};

#endif