///////////////////////////////////////////////////////////
// CSubTree.CPP
///////////////////////////////////////////////////////////
// Implementation of SubTree

#include "stdafx.h"
#include "CToken.h"
#include "CSubTree.h"


CSubTree::CSubTree()
{
	int x;
	for (x = 0; x < NODECHILDREN; x++)
		Children[x] = NULL;
	TheToken = new CToken;
}

CSubTree::CSubTree(TOKENTYPE InType, CString InVal)
{
	int x;
	for (x = 0; x < NODECHILDREN; x++)
		Children[x] = NULL;
	TheToken = new CToken;
	TheToken->SetType(InType);
	TheToken->SetVal(InVal);
}

CSubTree::~CSubTree()
{
	DeleteRecursive();
	if (TheToken != NULL)
		delete TheToken;
}

class CToken* CSubTree::GetToken()
{
 	return TheToken;
}


CSubTree* CSubTree::SetChild(int ChildIndex, CSubTree *Node)
{
	Children[ChildIndex] = Node;
	return Node;
}

CSubTree* CSubTree::SetNextChild(CSubTree * Node)
{
	return SetChild(GetChildCount(), Node);
}

bool CSubTree::IsEmpty()
{
	if (TheToken == NULL)
		return TRUE;
	return FALSE;
}


CSubTree* CSubTree::GetChild(int ChildIndex)
{
	return Children[ChildIndex];
}



void CSubTree::DeleteRecursive()
{
	int x;
	for (x = 0; x < NODECHILDREN; x++)
	{
		if (Children[x] != NULL)
		{
			Children[x]->DeleteRecursive();
			delete Children[x];
		}
		Children[x] = NULL;
	}
}

int CSubTree::GetChildCount()
{
	int x;
	for (x = 0; x < NODECHILDREN; x++)
	{
		if (Children[x] == NULL)
			break;
	}
	return x;
}

int CSubTree::GetDepth()
{
	return GetDepth(0);
}

int CSubTree::GetDepth(int CurrDepth)
{
	int index;
	int maxdepth;
	CurrDepth++;

	maxdepth = CurrDepth;

	for (index = 0; index < NODECHILDREN; index++)
	{
		if (Children[index] != NULL)
		{
			int ChildDepth;
			ChildDepth = Children[index]->GetDepth(CurrDepth);
			if (ChildDepth > maxdepth)
				maxdepth = ChildDepth;
		}
	}
	return maxdepth;
}

int CSubTree::GetNodeCount()
{
	int index;
	int CurrCount = 1;
	for (index = 0; index < NODECHILDREN; index++)
	{
		if (Children[index] != NULL)
			CurrCount = CurrCount + Children[index]->GetNodeCount();
	}
	return CurrCount;
}

// Recursive function:  Is at a node, pointing at the destination tree.
// Go through all children; if a child exists, create it, copy it to the
// destination tree, and recursively start again.
CSubTree * CSubTree::CreateCopy()
{
	CSubTree * DestTree;

	DestTree = new CSubTree;
	DestTree->GetToken()->SetType(GetToken()->GetType());
	DestTree->GetToken()->SetVal(GetToken()->GetVal());

	int index;
	for (index = 0; index < NODECHILDREN; index++)
	{
		if (Children[index] != NULL)
			// create child on DestTree
			DestTree->SetChild(index, Children[index]->CreateCopy());
	}

	return DestTree;
}

TOKENTYPE CSubTree::GetType()
{
	return GetToken()->GetType();
}

CString & CSubTree::GetVal()
{
	return GetToken()->GetVal();
}

void CSubTree::SetType(TOKENTYPE InType)
{
	GetToken()->SetType(InType);
}

void CSubTree::SetVal(CString InVal)
{
	GetToken()->SetVal(InVal);
}
