///////////////////////////////////////////////////////////
// CToken.CPP
///////////////////////////////////////////////////////////
// Contains the Token Class Methods:
// Constructor
// Destructor
// IsPrimitive() - Boolean determining whether the entire tree 
//                 consists of one node (i.e., an atomic function/value)
// IsTerminal()  - Boolean determining whether a specific token type
//                 is a terminal or non-terminal (may be unused by GAPS)
////////////////////////////////////////////////////////////

// Token Trees are built while scanning the input function string 
// left-to-right, and built according to standard precedence ideology.

// Primitive Token Trees are ones with only a single node; these 
// are associated solely with the operation of one single stored
// token itself.  To determine whether the item is only a primitive,
// a special boolean function, IsPrimitive() returns true if the 
// only node within the tree is a primitive.

// Example primitives are +, -, MOD, AND, etc.
// These primitive token trees are stored in the Standard "Delivered" Library.


#include "stdafx.h"
#include "afx.h"
#include "CToken.h"


CToken::CToken()
{
}

CToken::CToken(TOKENTYPE InType, CString & InTokenVal)
{
	SetType(InType);
	SetVal(InTokenVal);
}

CToken::~CToken()
{

}

void CToken::SetType(TOKENTYPE InType)
{
	TokenType = InType;
}

void CToken::SetVal(CString & InTokenVal)
{
	TokenVal = InTokenVal;
}

TOKENTYPE CToken::GetType()
{
	return TokenType;
}

CString & CToken::GetVal()
{
	return TokenVal;
}


