// DlgGAPSOutput.cpp : implementation file
//

#include "stdafx.h"
#include <fstream.h>
#include "GAPS2.h"
#include "CGAPSIndividual.h"
#include "EnvironmentDoc.h"
#include "DlgGAPSOutput.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDlgGAPSOutput dialog


CDlgGAPSOutput::CDlgGAPSOutput(CWnd* pParent /*=NULL*/)
	: CDialog(CDlgGAPSOutput::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDlgGAPSOutput)
	m_OutputText = _T("");
	//}}AFX_DATA_INIT

	m_HRMember = NULL;
	m_HREnvironment = NULL;
}


void CDlgGAPSOutput::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDlgGAPSOutput)
	DDX_Control(pDX, IDC_OUTPUTTREEVIEW, c_OutputTreeView);
	DDX_Control(pDX, IDC_VIEWTREE, m_ViewTree);
	DDX_Control(pDX, IDC_VIEWLIST, m_ViewList);
	DDX_Control(pDX, IDC_VIEWCCODE, m_ViewCode);
	DDX_Control(pDX, IDC_GAPSINDOUTPUT, m_OutputEditCtrl);
	DDX_Text(pDX, IDC_GAPSINDOUTPUT, m_OutputText);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CDlgGAPSOutput, CDialog)
	//{{AFX_MSG_MAP(CDlgGAPSOutput)
	ON_BN_CLICKED(IDC_VIEWLIST, OnViewlist)
	ON_BN_CLICKED(IDC_VIEWCCODE, OnViewccode)
	ON_BN_CLICKED(IDC_VIEWTREE, OnViewtree)
	ON_WM_CLOSE()
	ON_BN_CLICKED(IDC_GAPSSAVE, OnGapssave)
	ON_BN_CLICKED(IDC_QUICKIE, OnQuickie)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDlgGAPSOutput message handlers
void CDlgGAPSOutput::SetHRMember(CGAPSIndividual * HRMemberIn)
{
	m_HRMember = HRMemberIn;

	if (m_HREnvironment)
		m_OutputText = OutputCCode(m_HRMember->Tree);
	UpdateData(FALSE);


}

void CDlgGAPSOutput::SetHREnvironment(CEnvironmentDoc *HREnvIn)
{
	m_HREnvironment = HREnvIn;
	if (m_HRMember)
		m_OutputText = OutputCCode(m_HRMember->Tree);
	UpdateData(FALSE);

}


void CDlgGAPSOutput::OnViewlist() 
{
	m_ViewCode.SetCheck(MF_UNCHECKED);
	m_ViewTree.SetCheck(MF_CHECKED);
	m_ViewList.SetCheck(MF_UNCHECKED);

	m_OutputEditCtrl.ShowWindow(SW_SHOW);
	c_OutputTreeView.ShowWindow(SW_HIDE);

	m_OutputText = BuildStringFromTree(m_HRMember->Tree);
	UpdateData(FALSE);
}

void CDlgGAPSOutput::OnViewccode() 
{
	m_ViewCode.SetCheck(MF_CHECKED);
	m_ViewTree.SetCheck(MF_UNCHECKED);
	m_ViewList.SetCheck(MF_UNCHECKED);

	m_OutputEditCtrl.ShowWindow(SW_SHOW);
	c_OutputTreeView.ShowWindow(SW_HIDE);

	m_OutputText = OutputCCode(m_HRMember->Tree);
	UpdateData(FALSE);
}

void CDlgGAPSOutput::OnViewtree() 
{
	m_ViewCode.SetCheck(MF_UNCHECKED);
	m_ViewTree.SetCheck(MF_CHECKED);
	m_ViewList.SetCheck(MF_UNCHECKED);

	m_OutputEditCtrl.ShowWindow(SW_HIDE);
	c_OutputTreeView.ShowWindow(SW_SHOW);

	FillTree();
	UpdateData(FALSE);
}


void CDlgGAPSOutput::OnClose() 
{	
	CDialog::OnClose();
	delete this;
}

void CDlgGAPSOutput::OnGapssave() 
{
	// Save the current output text to a file
	static char BASED_CODE szFilter[] = "C source code files(*.c)|*.c";

	// Create File Save As Dialog
	CFileDialog OurFileDlg (
		FALSE,                // Save As
		_T(".c"),             // Default C-code file extension
		_T("GAPSGenFunct"),       // Default GAPS 2 Library File Name
		OFN_OVERWRITEPROMPT,  // If file exists, ask if user wants to over-write
		szFilter);

	if (OurFileDlg.DoModal() == IDOK)
	{
		ofstream OutFile;
		OutFile.open(OurFileDlg.GetPathName());
		OutFile << m_OutputText << endl;
		OutFile.close();
	}
}

BOOL CDlgGAPSOutput::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	m_ViewCode.SetCheck(MF_CHECKED);
	m_ViewTree.SetCheck(MF_UNCHECKED);
	m_ViewList.SetCheck(MF_UNCHECKED);
	

	// Set OutputWindow font to Courier
	CFont OutputFont;
	OutputFont.CreateFont(16,0, 0, 0, FW_NORMAL, 0, 0, 0,
		DEFAULT_CHARSET, OUT_RASTER_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, 
		DEFAULT_PITCH, "Courier New");	
	
	m_OutputEditCtrl.SetFont(&OutputFont);

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}


// View as C-Code (Default)
CString CDlgGAPSOutput::OutputCCode(CSubTree * NodeIn)
{
	CString GeneratedCCode;
	CString FormatIt;
	int index;
	int parmindex;
	CString RPBCode;

	GeneratedCCode  = CString("/************************************************************/\n");
	GeneratedCCode += CString("/* Code generated by the GP-Lab Genetic Programming System. */\n");
	GeneratedCCode += CString("/* All Attributes are Global constructs                     */\n");
	GeneratedCCode += CString("/************************************************************/\n");
	GeneratedCCode += CString("/* Note: This system assumes the use of the MFC 'CString'   */\n");
	GeneratedCCode += CString("/* class for the String implementation.  Many excellent     */\n");
	GeneratedCCode += CString("/* CString Clones can be found through a quick internet     */\n");
	GeneratedCCode += CString("/* search on the phrase 'CString Clone'                     */\n");
	GeneratedCCode += CString("/************************************************************/\n");
	GeneratedCCode += CString("/* Also, a few of GP-Lab's language statements require the  */\n");
	GeneratedCCode += CString("/* use of some implementation of a stack data construct.    */\n");
	GeneratedCCode += CString("/* The implementations of Push and Pop in GP-Lab are synth- */\n");
	GeneratedCCode += CString("/* esized; it is recommended that the programmer build his  */\n");
	GeneratedCCode += CString("/* or her own versions of a Stack as the need fits.         */\n");
	GeneratedCCode += CString("/************************************************************/\n");

	CGPProgram OurFullProgram;

	OurFullProgram.ProgramTree = m_HREnvironment->t_MainProgramTree->ProgramTree->CreateCopy();

	// Create Global Definitions
	for (index = 0 ; index < m_HRMember->AttributeTable->GetAttributeCount(); index++)
	{
		switch (m_HRMember->AttributeTable->GetAttribType(m_HRMember->AttributeTable->GetAttribNameByNumber(index)))
		{
		case rtFLOAT:
			GeneratedCCode += CString("double ");
			break;
		case rtBOOL:
			GeneratedCCode += CString("int ");
			break;
		case rtSTRING:
			GeneratedCCode += CString("CString ");
			break;
		}
		GeneratedCCode += m_HRMember->AttributeTable->GetAttribNameByNumber(index);
		GeneratedCCode += CString(";\n");
	}

	// Build all functions, user-created as well as GP-Lab Specific
	for (index = 0; index < m_HREnvironment->m_TotalUserEnteredFunctions; index++)
	{
		GeneratedCCode += CString("\n\n");
		GeneratedCCode += CString("/************************************************************/\n");
		GeneratedCCode += CString("  ");
		switch (m_HREnvironment->tbl_OurFunctionTable.GetType(m_HREnvironment->a_UserEnteredFunctions[index].Name))
		{
		case rtFLOAT:
			GeneratedCCode += CString("double ");
			break;
		case rtBOOL:
			GeneratedCCode += CString("int ");
			break;
		case rtSTRING:
			GeneratedCCode += CString("CString ");
			break;
		case rtVOID:
			GeneratedCCode += CString("void ");
			break;
		}
		GeneratedCCode += m_HREnvironment->a_UserEnteredFunctions[index].Name;
		GeneratedCCode += CString ("(");
		for (parmindex = 0; parmindex < m_HREnvironment->tbl_OurFunctionTable.GetParameterCount(m_HREnvironment->a_UserEnteredFunctions[index].Name); parmindex++)
		{
			if (parmindex > 0)
				GeneratedCCode += CString(", ");
			switch (m_HREnvironment->tbl_OurFunctionTable.GetParmType(m_HREnvironment->a_UserEnteredFunctions[index].Name, parmindex))
			{
				case rtFLOAT:
					GeneratedCCode += CString("double ");
					break;
				case rtBOOL:
					GeneratedCCode += CString("int ");
					break;
				case rtSTRING:
					GeneratedCCode += CString("CString ");
					break;
			}

			GeneratedCCode += CString(" ");
			GeneratedCCode += m_HREnvironment->tbl_OurFunctionTable.GetParmName(m_HREnvironment->a_UserEnteredFunctions[index].Name, parmindex);
		}


		GeneratedCCode += CString(")\n");

		GeneratedCCode += CString("/************************************************************/\n");
		GeneratedCCode += CString("{\n\n\t");
		switch (m_HREnvironment->tbl_OurFunctionTable.GetType(m_HREnvironment->a_UserEnteredFunctions[index].Name))
		{
		case rtFLOAT:
			GeneratedCCode += CString("double ") + m_HREnvironment->a_UserEnteredFunctions[index].Name + CString("\n\n");
			break;
		case rtBOOL:
			GeneratedCCode += CString("int ") + m_HREnvironment->a_UserEnteredFunctions[index].Name + CString("\n\n");
			break;
		case rtSTRING:
			GeneratedCCode += CString("CString ") + m_HREnvironment->a_UserEnteredFunctions[index].Name + CString("\n\n");
			break;
		}		

		FormatIt = m_HREnvironment->a_UserEnteredFunctions[index].Name;
		GeneratedCCode += GenCodeFromTree(m_HREnvironment->tbl_OurFunctionTable.GetFunctionProgram(FormatIt)->ProgramTree, 1);

		if (m_HREnvironment->tbl_OurFunctionTable.GetType(m_HREnvironment->a_UserEnteredFunctions[index].Name) != rtVOID)
		{
			GeneratedCCode += CString("\treturn ") + m_HREnvironment->a_UserEnteredFunctions[index].Name + CString(";");
		}
		GeneratedCCode += CString("\n}\n\n");
	}

	RPBCode = GenCodeFromTree(OurFullProgram.ProgramTree, 1);

	if (GeneratedCCode.Find("GetAt") > -1 ||  RPBCode.Find("GetAt") > -1)
	{

		GeneratedCCode += CString("/************************************************************/\n");
		GeneratedCCode += CString("CString GetAt(CString Source, int Start, int Length)\n");
		GeneratedCCode += CString("{\n");
		GeneratedCCode += CString("   return Source.Mid(Start, Length);\n");
		GeneratedCCode += CString("}\n");
		GeneratedCCode += CString("/************************************************************/\n\n");	
	}

	if (GeneratedCCode.Find("Replace") > -1  || RPBCode.Find("Replace") > -1)
	{
		GeneratedCCode += CString("/************************************************************/\n");
		GeneratedCCode += CString("void Replace(CString & StrAttribute, CString Source, int Start, int Length)\n");
		GeneratedCCode += CString("{\n");
		GeneratedCCode += CString("   CString OldRight;\n\n");
		GeneratedCCode += CString("   OldRight = StrAttribute.Right(StrAttribute.GetLength() - Start);\n");
		GeneratedCCode += CString("   StrAttribute = StrAttribute.Left(Start)\n");
		GeneratedCCode += CString("   if (Source.GetLength() > OldRight.GetLength())\n");
		GeneratedCCode += CString("   {\n");
		GeneratedCCode += CString("      StrAttribute += Source;\n");
		GeneratedCCode += CString("   }\n");
		GeneratedCCode += CString("   else\n");
		GeneratedCCode += CString("   {\n");
		GeneratedCCode += CString("      StrAttribute += Source + OldRight.Right(OldRight.GetLength() - Length);\n");
		GeneratedCCode += CString("   }\n");
		GeneratedCCode += CString("}\n");
		GeneratedCCode += CString("/************************************************************/\n\n");	
	}	

	// Build 'main'
	GeneratedCCode += CString("\n");
	GeneratedCCode += CString("/************************************************************/\n");
	GeneratedCCode += CString("/* main                                                     */\n");
	GeneratedCCode += CString("/************************************************************/\n\n");
	GeneratedCCode += CString("void main(int argc, char argv[])\n");
	GeneratedCCode += CString("{\n");

	GeneratedCCode += RPBCode;

	GeneratedCCode += CString("\n}\n");

	

	return GeneratedCCode;
}

CString CDlgGAPSOutput::GenCodeFromTree(CSubTree * NodeIn, int TabDepth)
{

	CString ReturnVal;
	int index;
	CString FormatIt;
	CString TabsIn = CString("");

	for (index = 0; index < TabDepth; index++)
		TabsIn += CString("\t");

	switch (NodeIn->GetToken()->GetType())
	{

	case ttGENETICFUNC:
		return GenCodeFromTree(m_HRMember->Tree, TabDepth);

	case ttUSRATTRIB:
	case ttFLOAT:
	case ttBOOL:
		return NodeIn->GetVal();

	case ttSTRING:
		ReturnVal.Format("CString(\"%s\")", NodeIn->GetVal());
		return ReturnVal;

	case ttUSRFUNCTION:
		ReturnVal.Format("%s%s", m_HREnvironment->tbl_OurFunctionTable.GetType(NodeIn->GetVal()) == rtVOID ? TabsIn : CString(""), NodeIn->GetVal());
		ReturnVal += CString ("(");
		for (index = 0; index < m_HREnvironment->tbl_OurFunctionTable.GetParameterCount(NodeIn->GetVal()); index++)
		{
			if (index > 0)
				ReturnVal += CString (", ");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(index), TabDepth);
		}
		ReturnVal += CString(")");
		if (m_HREnvironment->tbl_OurFunctionTable.GetType(NodeIn->GetVal()) == rtVOID)
			ReturnVal += CString(";\n");

		if (NodeIn->GetChild(index) != NULL)
		{
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(index), TabDepth);
		}

		return ReturnVal;

	case ttNEGATIVE:
		ReturnVal.Format("(-1 * (%s))", GenCodeFromTree(NodeIn->GetChild(0), TabDepth));
		return ReturnVal;

	case ttNOT:
		ReturnVal.Format("(!(%s))", GenCodeFromTree(NodeIn->GetChild(0), TabDepth));
		return ReturnVal;

	case ttAND:
		ReturnVal.Format("%s && %s", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttOR:
		ReturnVal.Format("(%s || %s)", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttCOMPAREOP:
		ReturnVal.Format("(%s %s %s)", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), NodeIn->GetVal(), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;
		
	case ttADD:
		ReturnVal.Format("(%s + %s)", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttSUBTRACT:
		ReturnVal.Format("(%s - %s)", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttMULTIPLY:
		ReturnVal.Format("%s * %s", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttDIVIDE:
		ReturnVal.Format("%s / %s", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttMOD:
		ReturnVal.Format("%s %c %s", GenCodeFromTree(NodeIn->GetChild(0), TabDepth), '%', GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		return ReturnVal;

	case ttFNGETAT:
		ReturnVal.Format("GetAt(%s, %s, %s)", NodeIn->GetChild(0)->GetVal(),
			GenCodeFromTree(NodeIn->GetChild(1), TabDepth), 
			GenCodeFromTree(NodeIn->GetChild(2), TabDepth));
		return ReturnVal;

	// Special cases: Built-in statements
	case ttFNASSIGN:
		ReturnVal.Format("%s%s = %s;\n", TabsIn, NodeIn->GetChild(0)->GetVal(), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNIF:
		ReturnVal.Format("%sif (%s)\n", TabsIn, GenCodeFromTree(NodeIn->GetChild(0), TabDepth));
		ReturnVal += TabsIn + CString("{\n");
		ReturnVal += GenCodeFromTree(NodeIn->GetChild(1), TabDepth+1);
		ReturnVal += TabsIn + CString("}\n");
		if (NodeIn->GetChild(2)->GetType() != ttFNNOOP)
		{
			ReturnVal += TabsIn + CString("else\n");
			ReturnVal += TabsIn + CString("{\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth+1);
			ReturnVal += TabsIn + CString("}\n");
		}
		if (NodeIn->GetChildCount() > 3)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(3), TabDepth);
		}
		return ReturnVal;

	case ttFNWHILE:
		ReturnVal.Format("%swhile (%s)\n", TabsIn, GenCodeFromTree(NodeIn->GetChild(0), TabDepth));
		ReturnVal += TabsIn + CString("{\n");
		ReturnVal += GenCodeFromTree(NodeIn->GetChild(1), TabDepth+1);
		ReturnVal += TabsIn + CString("}\n");
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNDOUNTIL:
		ReturnVal = TabsIn + CString("do\n");
		ReturnVal += TabsIn + CString("{\n");
		ReturnVal += GenCodeFromTree(NodeIn->GetChild(0), TabDepth+1);
		ReturnVal.Format("%s} while (%s)\n", TabsIn, GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNPUSH:
		ReturnVal.Format("%sPush(%s, %s);\n", TabsIn, NodeIn->GetChild(0)->GetVal(), GenCodeFromTree(NodeIn->GetChild(1), TabDepth));
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNPOP:
		ReturnVal.Format("%sPop(%s, %s);\n", TabsIn, NodeIn->GetChild(0)->GetVal(), NodeIn->GetChild(1)->GetVal());
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNREPLACE:
		ReturnVal.Format("%sReplace(%s, %s, %s, %s);\n", TabsIn, 
			NodeIn->GetChild(0)->GetVal(), 
			GenCodeFromTree(NodeIn->GetChild(1), TabDepth), 
			GenCodeFromTree(NodeIn->GetChild(2), TabDepth), 
			GenCodeFromTree(NodeIn->GetChild(3), TabDepth));
		if (NodeIn->GetChildCount() > 4)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(4), TabDepth);
		}
		return ReturnVal;

	case ttFNMAP:
		ReturnVal.Format("%sswitch (%s)\n", TabsIn, NodeIn->GetChild(0)->GetVal());
		ReturnVal += TabsIn + CString("{\n");
		ReturnVal += GenCodeFromTree(NodeIn->GetChild(1), TabDepth + 1);
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth+1);
		}
		ReturnVal += TabsIn + CString("}\n");
		return ReturnVal;

	case ttFNVALUE:
		ReturnVal.Format("\n%scase %s:\n", TabsIn, GenCodeFromTree(NodeIn->GetChild(0), TabDepth));
		ReturnVal += TabsIn + CString("{\n");
		ReturnVal += TabsIn + GenCodeFromTree(NodeIn->GetChild(1), TabDepth+1);
		ReturnVal += TabsIn + CString("}\n");
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += CString("\n");
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNDECLARE:
		if (NodeIn->GetChild(1)->GetVal() == CString("Number"))
			ReturnVal.Format("%sdouble %s;\n", TabsIn, NodeIn->GetChild(0)->GetVal());
		if (NodeIn->GetChild(1)->GetVal() == CString("Boolean"))
			ReturnVal.Format("%sint %s;\n", TabsIn, NodeIn->GetChild(0)->GetVal());
		if (NodeIn->GetChild(1)->GetVal() == CString("String"))
			ReturnVal.Format("%sCString %s;\n", TabsIn, NodeIn->GetChild(0)->GetVal());
		if (NodeIn->GetChildCount() > 2)
		{
			ReturnVal += GenCodeFromTree(NodeIn->GetChild(2), TabDepth);
		}
		return ReturnVal;

	case ttFNNOOP:
		return CString("");

	}

	
	return CString("<*ERROR*>");
}

/////////////////////////////////////////////////////////////////////////
// View as lisp-style list
CString CDlgGAPSOutput::BuildStringFromTree(CSubTree *NodeIn)
{
	CString RetVal;
	int index;

	if (NodeIn->GetType() == ttFLOAT ||
		NodeIn->GetType() == ttBOOL ||
		NodeIn->GetType() == ttSTRING ||
		NodeIn->GetType() == ttUSRATTRIB)
	{
		return NodeIn->GetVal();
	}

	RetVal.Format(" %s(", NodeIn->GetVal());
	for (index = 0; index < NodeIn->GetChildCount(); index++)
	{
		if (index > 0)
			RetVal += CString(", ");
		RetVal = RetVal + BuildStringFromTree(NodeIn->GetChild(index));
	}
	RetVal += CString(")");
	return RetVal;
}


/////////////////////////////////////////////////////////////////////////
// View as graphical Tree 
void CDlgGAPSOutput::FillTree()
{
	HTREEITEM RootNode;
	CString FormatIt;
	int index;

	c_OutputTreeView.DeleteAllItems();

	// Create root Main Program node
	RootNode = c_OutputTreeView.InsertItem("Main Program Tree");
	BuildTreeNode(RootNode, m_HREnvironment->t_MainProgramTree->ProgramTree);


	// Create root Result Branch
	RootNode = c_OutputTreeView.InsertItem("Result Branch");
	BuildTreeNode(RootNode, m_HRMember->Tree);

	// Create each function 
	for (index = 0; index < m_HREnvironment->m_TotalUserEnteredFunctions; index++)
	{
		FormatIt.Format("User Function %s", m_HREnvironment->a_UserEnteredFunctions[index].Name);
		RootNode = c_OutputTreeView.InsertItem(FormatIt);
		BuildTreeNode(RootNode, m_HREnvironment->tbl_OurFunctionTable.GetFunctionProgram(m_HREnvironment->a_UserEnteredFunctions[index].Name)->ProgramTree);
	}
}

void CDlgGAPSOutput::BuildTreeNode(HTREEITEM Parent, CSubTree *NodeIn)
{
	HTREEITEM NewParent;
	int index;

	NewParent = c_OutputTreeView.InsertItem(NodeIn->GetVal(), Parent);

	for (index = 0; index < NodeIn->GetChildCount(); index++)
		BuildTreeNode(NewParent, NodeIn->GetChild(index));

}

void CDlgGAPSOutput::OnQuickie() 
{
	struct ArgsAndTargets *ThisArgTargPair;
	CString FormatIt;
	CGPProgram GPLabExecution;
	double CurrFitness = 0.0;

	double FinalFitness = 0.0;

	m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg1Type, CString("Argument0"), 0);
	if (m_HREnvironment->m_NumArguments > 1)
		m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg2Type, CString("Argument1"), 0);
	if (m_HREnvironment->m_NumArguments > 2)
		m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg3Type, CString("Argument2"), 0);
	if (m_HREnvironment->m_NumArguments > 3)
		m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg4Type, CString("Argument3"), 0);
	if (m_HREnvironment->m_NumArguments > 4)
		m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg5Type, CString("Argument4"), 0);
	if (m_HREnvironment->m_NumArguments > 5)
		m_HRMember->AttributeTable->AddAttribute(m_HREnvironment->tbl_OurTargets->Arg6Type, CString("Argument5"), 0);
	
	// Add Specialized attribute "Target"
	m_HRMember->AttributeTable->AddAttribute(rtFLOAT, "Target", 0);

	for (ThisArgTargPair = m_HREnvironment->tbl_OurTargets; ThisArgTargPair != NULL; ThisArgTargPair = ThisArgTargPair->next)
	{
		m_HRMember->AttributeTable->SetAttrib(CString("Argument0"), ThisArgTargPair->Arg1);
		if (m_HREnvironment->m_NumArguments > 1)
			m_HRMember->AttributeTable->SetAttrib(CString("Argument1"), ThisArgTargPair->Arg2);
		if (m_HREnvironment->m_NumArguments > 2)
			m_HRMember->AttributeTable->SetAttrib(CString("Argument2"), ThisArgTargPair->Arg3);
		if (m_HREnvironment->m_NumArguments > 3)
			m_HRMember->AttributeTable->SetAttrib(CString("Argument3"), ThisArgTargPair->Arg4);
		if (m_HREnvironment->m_NumArguments > 4)
			m_HRMember->AttributeTable->SetAttrib(CString("Argument4"), ThisArgTargPair->Arg5);
		if (m_HREnvironment->m_NumArguments > 5)
			m_HRMember->AttributeTable->SetAttrib(CString("Argument5"), ThisArgTargPair->Arg6);

		
		GPLabExecution.ProgramTree = m_HREnvironment->t_MainProgramTree->ProgramTree;
		GPLabExecution.SetRPBTree(m_HRMember->Tree);

		GPLabExecution.Execute(m_HRMember->AttributeTable, &m_HREnvironment->tbl_OurFunctionTable, m_HREnvironment->m_AbsoluteTimeLimit);

		m_HRMember->AttributeTable->SetAttrib(CString("Target"), ThisArgTargPair->TargetVal);
		CurrFitness += atof(m_HREnvironment->t_RawFitnessTree->Execute(m_HRMember->AttributeTable, &m_HREnvironment->tbl_OurFunctionTable, 0));
		GPLabExecution.ProgramTree = NULL;
	}

	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument0"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument1"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument2"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument3"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument4"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Argument5"));
	m_HRMember->AttributeTable->RemoveAttribute(CString("Target"));

	FormatIt.Format("%f", CurrFitness);


}		

