// EnvironmentDoc.cpp : implementation file
//

#include "stdafx.h"
#include "GAPS2.h"
#include "EnvironmentDoc.h"
#include "CGAPSExecution.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

bool IsNumber(CString & TestStr );

/////////////////////////////////////////////////////////////////////////////
// CEnvironmentDoc

IMPLEMENT_DYNCREATE(CEnvironmentDoc, CDocument)

CEnvironmentDoc::CEnvironmentDoc()
{
	int index;
	for (index = 0; index < TOTALBUILTINS; index++)
	{
		a_SelectedBuiltIns[index] = CString("");
	}

	m_FileLoaded = FALSE;
	t_MainProgramTree  = FALSE;
	m_TotalSelectedBuiltIns = 0;
	t_RawFitnessTree = NULL;
	t_StdFitnessTree = NULL;

}

BOOL CEnvironmentDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	tbl_OurTargets = NULL;

	return TRUE;
}

CEnvironmentDoc::~CEnvironmentDoc()
{

	// Delete Aguments/Targets
	if (tbl_OurTargets != NULL)
	{
		struct ArgsAndTargets *CurrNode;
		struct ArgsAndTargets *DelNode;
		for (CurrNode = tbl_OurTargets; CurrNode != NULL;)
		{
			DelNode = CurrNode;
			CurrNode = CurrNode->next;
			delete DelNode;
		}
	}

	ResetTrees();

}

void CEnvironmentDoc::ResetTrees()
{

	t_RPBNode.ProgramTree = NULL;

	if (t_MainProgramTree != NULL)
		delete t_MainProgramTree;
	t_MainProgramTree = NULL;

	if (t_RawFitnessTree != NULL)
		delete t_RawFitnessTree;
	t_RawFitnessTree = NULL;

	if (t_StdFitnessTree != NULL)
		delete t_StdFitnessTree;
	t_StdFitnessTree = NULL;

	tbl_OurFunctionTable.ClearAll();
}


BEGIN_MESSAGE_MAP(CEnvironmentDoc, CDocument)
	//{{AFX_MSG_MAP(CEnvironmentDoc)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEnvironmentDoc diagnostics

#ifdef _DEBUG
void CEnvironmentDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CEnvironmentDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CEnvironmentDoc serialization

void CEnvironmentDoc::Serialize(CArchive& ar)
{
	WORD GetVal;

	 // Environmental Save Sentinels
	 // es_ENDENVIRONMENT, es_ENVIRONNAME, es_ENVIRONDESCR
	 // es_MUTATIONRATE, es_MAXUNPRODUCTIVEGENS, es_SHRINKAGEGENS,
	 // es_POPULATION, es_HONORROLLSIZE, es_ABSOLUTETIMELIMIT,
	 // es_FITNESSTHRESHOLD, es_GENERATIONSTRATEGY, es_MINIMUMDEPTH,
	 // es_MAXIMUMDEPTH, es_SELECTIONSTRATEGY, es_SELECTIONPARAMETER,
	 // es_TOURNAMENTSIZE, es_NUMBEROFARGUMENTS, 
	 // es_NUMBEROFTARGETS, es_TARGETLIST, es_USERFUNCTIONS, es_NUMBEROFUSRFUNCTS,
	 // es_BASEPROGRAM, es_RAWFITNESSFORMULA, es_STANDARDFITNESSFORMULA, 
	 // es_SELECTEDBUILTINS, es_TOTALATTRIBUTES, es_ATTRIBUTES, 
	 // es_SELECTIONATTRIBUTES, es_SELECTIONRANDONNUMS, es_SELECTIONNUMDECIMALS,
	 // es_SELECTIONNUMLOW, es_SELECTIONNUMHIGH, es_SELECTIONBOOLEANS,
	 // es_SELECTIONSTRINGS, es_SELECTIONCHARSLOW, es_SELECTIONCHARSHIGH,
	 // es_SELECTIONSTRLENLOW, es_SELECTIONSTRLENHIGH, es_SELECTIONARGS

	int index;
	CObject::Serialize( ar );
	if (ar.IsStoring())
	{
		// save Identification
		ar << (WORD) es_ENVIRONNAME;
		ar << m_EnvironName;

		ar << (WORD) es_ENVIRONDESCR;
		ar << m_EnvironDescr;

		ar << (WORD) es_MUTATIONRATE;
		ar << m_MutationRate;

		// save Environmental Parameters
		ar << (WORD) es_MAXUNPRODUCTIVEGENS;
		ar << m_MaxUnproductiveGens;

		ar << (WORD) es_SHRINKAGEGENS;
		ar << m_ShrinkageGenerations;

		ar << (WORD) es_POPULATION;
		ar << m_Population;

		ar << (WORD) es_HONORROLLSIZE;
		ar << m_HonorRollSize;

		ar << (WORD) es_ABSOLUTETIMELIMIT;
		ar << m_AbsoluteTimeLimit;

		ar << (WORD) es_FITNESSTHRESHOLD;
		ar << m_FitnessThreshold;

		ar << (WORD) es_GENERATIONSTRATEGY;
		ar << m_GenerationStrategy;

		ar << (WORD) es_MAXIMUMDEPTH;
		ar << m_MaximumDepth;

		ar << (WORD) es_MINIMUMDEPTH;
		ar << m_MinimumDepth;

		ar << (WORD) es_SELECTIONSTRATEGY;
		ar << m_SelectionStrategy;

		ar << (WORD) es_SELECTIONPARAMETER;
		ar << m_SelectionParmNum;
	
		ar << (WORD) es_TOURNAMENTSIZE;
		ar << m_TournamentSize;

		// save Arguments and Targets
		ar << (WORD) es_NUMBEROFARGUMENTS;
		ar << m_NumArguments;

		ar << (WORD) es_NUMBEROFTARGETS;
		ar << m_NumTargets;

		ar << (WORD) es_TARGETLIST;
		struct ArgsAndTargets *CurrTarget;
		for (CurrTarget = tbl_OurTargets; CurrTarget != NULL; CurrTarget = CurrTarget->next)
		{
			ar << CurrTarget->Arg1;
			ar << CurrTarget->Arg2;
			ar << CurrTarget->Arg3;
			ar << CurrTarget->Arg4;
			ar << CurrTarget->Arg5;
			ar << CurrTarget->Arg6;
			ar << CurrTarget->TargetVal;
			ar << (WORD) CurrTarget->Arg1Type;
			ar << (WORD) CurrTarget->Arg2Type;
			ar << (WORD) CurrTarget->Arg3Type;
			ar << (WORD) CurrTarget->Arg4Type;
			ar << (WORD) CurrTarget->Arg5Type;
			ar << (WORD) CurrTarget->Arg6Type;
			ar << (WORD) CurrTarget->TargetType;
		}

		// save User Functions
		ar << (WORD) es_NUMBEROFUSRFUNCTS;
		ar << m_TotalUserEnteredFunctions;

		ar << (WORD) es_USERFUNCTIONS;
		for (index = 0; index < m_TotalUserEnteredFunctions; index++)
		{
			ar << a_UserEnteredFunctions[index].Name;
			ar << a_UserEnteredFunctions[index].Type;
			GetVal = a_UserEnteredFunctions[index].Selected;
			ar << GetVal;
			ar << a_UserEnteredFunctions[index].Text;
			ar << a_UserEnteredFunctions[index].TotalArguments;

			int argindex;
			for (argindex = 0; argindex < a_UserEnteredFunctions[index].TotalArguments; argindex++)
			{
				ar << a_UserEnteredFunctions[index].Arguments[argindex];
				ar << a_UserEnteredFunctions[index].ArgumentTypes[argindex];
			}
		}
		// save Selected Builtins
		ar << (WORD) es_SELECTEDBUILTINS;
		ar << m_TotalSelectedBuiltIns;
		for (index = 0; index < TOTALBUILTINS; index++)
		{
			ar << a_SelectedBuiltIns[index];
		}

		// save Base Program
		ar << (WORD) es_BASEPROGRAM;
		ar << m_BaseProgram;

		// save Fitness Calculation
		ar << (WORD) es_RAWFITNESSFORMULA;
		ar << m_RawFitnessFormula;

		ar << (WORD) es_STANDARDFITNESSFORMULA;
		ar << m_StandardizedFitnessFormula;

		// save Attributes
		ar << (WORD) es_TOTALATTRIBUTES;
		ar << m_TotalAttributes;

		ar << (WORD) es_ATTRIBUTES;
		for (index = 0; index < m_TotalAttributes; index++)
		{
			ar << a_Attributes[index].Name;
			ar << a_Attributes[index].Type;
			ar << (WORD) a_Attributes[index].IsPersistent;
			ar << (WORD) a_Attributes[index].IsSelectable;
		}

		// Save Terminals Selected
		ar << (WORD) es_SELECTIONATTRIBUTES;
		GetVal = m_TermsSelectedAttributes;
		ar << (WORD) GetVal;

		ar << (WORD) es_SELECTIONRANDONNUMS;
		GetVal = m_TermsSelectedRandNums;
		ar << (WORD) GetVal;

		ar << (WORD) es_SELECTIONNUMDECIMALS;
		ar << m_TermsSelectedNumDecimals;

		ar << (WORD) es_SELECTIONNUMLOW;
		ar << m_TermsSelectedNumLow;

		ar << (WORD) es_SELECTIONNUMHIGH;
		ar << m_TermsSelectedNumHigh;

		ar << (WORD) es_SELECTIONBOOLEANS;
		GetVal = m_TermsSelectedBools;
		ar << (WORD) GetVal;

		ar << (WORD) es_SELECTIONSTRINGS;
		GetVal = m_TermsSelectedStrings;
		ar << (WORD) GetVal;

		ar << (WORD) es_SELECTIONCHARSLOW;
		ar << m_TermsSelectedCharLow;

		ar << (WORD) es_SELECTIONCHARSHIGH;
		ar << m_TermsSelectedCharHigh;

		ar << (WORD) es_SELECTIONSTRLENLOW;
		ar << m_TermsSelectedCharLenLow;

		ar << (WORD) es_SELECTIONSTRLENHIGH;
		ar << m_TermsSelectedCharLenHigh;

		ar << (WORD) es_SELECTIONARGS;
		GetVal = m_TermsSelectedArgs;
		ar << (WORD) GetVal;

		// End-Of-Environment Sentinel
		ar << (WORD) es_ENDENVIRONMENT;

		m_FileLoaded = FALSE;

	}
	else // LOADING
	{
		WORD ParmCommand;
		WORD TempVal;
		BOOL ContinueExec = TRUE;

		while (ContinueExec)
		{
			ar >> ParmCommand;
			switch ((enum ENVIRONSER ) ParmCommand)
			{
			case es_ENVIRONNAME:
				ar >> m_EnvironName;
				break;
		
			case es_ENVIRONDESCR:
				ar >> m_EnvironDescr;
				break;
			
			case es_MUTATIONRATE:
				ar >> m_MutationRate;
				break;
			case es_MAXUNPRODUCTIVEGENS:
				ar >> m_MaxUnproductiveGens;
				break;
			case es_SHRINKAGEGENS:
				ar >> m_ShrinkageGenerations;
				break;
			case es_POPULATION:
				ar >> m_Population;
				break;
			case es_HONORROLLSIZE:
				ar >> m_HonorRollSize;
				break;
			case es_ABSOLUTETIMELIMIT:
				ar >> m_AbsoluteTimeLimit;
				break;
			case es_FITNESSTHRESHOLD:
				ar >> m_FitnessThreshold;
				break;
			case es_GENERATIONSTRATEGY:
				ar >> m_GenerationStrategy;
				break;
			case es_MAXIMUMDEPTH:
				ar >> m_MaximumDepth;
				break;
			case es_MINIMUMDEPTH:
				ar >> m_MinimumDepth;
				break;
			case es_SELECTIONSTRATEGY:
				ar >> m_SelectionStrategy;
				break;
			case es_SELECTIONPARAMETER:
				ar >> m_SelectionParmNum;
				break;
			case es_TOURNAMENTSIZE:
				ar >> m_TournamentSize;
				break;
			case es_NUMBEROFARGUMENTS:
				ar >> m_NumArguments;
				break;
			case es_NUMBEROFTARGETS:
				ar >> m_NumTargets;
				break;
			case es_TARGETLIST:
				tbl_OurTargets = NULL;
				if (m_NumTargets > 0)
				{
					struct ArgsAndTargets *CurrTarget;
					for (index = 0; index < m_NumTargets; index++)
					{
						if (index==0)
							CurrTarget = tbl_OurTargets = new struct ArgsAndTargets;
						else
						{
							CurrTarget->next = new struct ArgsAndTargets;
							CurrTarget = CurrTarget->next;
						}
						CurrTarget->next = NULL;
						ar >> CurrTarget->Arg1;
						ar >> CurrTarget->Arg2;
						ar >> CurrTarget->Arg3;
						ar >> CurrTarget->Arg4;
						ar >> CurrTarget->Arg5;
						ar >> CurrTarget->Arg6;
						ar >> CurrTarget->TargetVal;
						ar >> TempVal;
						CurrTarget->Arg1Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->Arg2Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->Arg3Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->Arg4Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->Arg5Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->Arg6Type = (enum RETURNTYPE) TempVal;
						ar >> TempVal;
						CurrTarget->TargetType = (enum RETURNTYPE) TempVal;
					} // for index 
				} // m_NumTargets > 0
				break;

			case es_NUMBEROFUSRFUNCTS:
				ar >> m_TotalUserEnteredFunctions;
				break;
			case es_USERFUNCTIONS:
				for (index = 0; index < m_TotalUserEnteredFunctions; index++)
				{
					ar >> a_UserEnteredFunctions[index].Name;
					ar >> a_UserEnteredFunctions[index].Type;
					ar >> TempVal;
					a_UserEnteredFunctions[index].Selected = (bool) (TempVal & 0x01);
					ar >> a_UserEnteredFunctions[index].Text;
					ar >> a_UserEnteredFunctions[index].TotalArguments;

					int argindex;
					for (argindex = 0; argindex < a_UserEnteredFunctions[index].TotalArguments; argindex++)
					{
						ar >> a_UserEnteredFunctions[index].Arguments[argindex];
						ar >> a_UserEnteredFunctions[index].ArgumentTypes[argindex];
					}
				}
				break;
			case es_SELECTEDBUILTINS:
				ar >> m_TotalSelectedBuiltIns;
				for (index = 0; index < TOTALBUILTINS; index++)
				{
					ar >> a_SelectedBuiltIns[index];
				}
				break;
			case es_BASEPROGRAM:
				ar >> m_BaseProgram;
				break;
			case es_RAWFITNESSFORMULA:
				ar >> m_RawFitnessFormula;
				break;
			case es_STANDARDFITNESSFORMULA:
				ar >> m_StandardizedFitnessFormula;
				break;
			case es_TOTALATTRIBUTES:
				ar >> m_TotalAttributes;
				break;
			case es_ATTRIBUTES:
				for (index = 0; index < m_TotalAttributes; index++)
				{
					ar >> a_Attributes[index].Name;
					ar >> a_Attributes[index].Type;
					ar >> TempVal;
					a_Attributes[index].IsPersistent = (bool) (TempVal & 0x01);
					ar >> TempVal;
					a_Attributes[index].IsSelectable = (bool) (TempVal & 0x01);
				}
				break;
			case es_SELECTIONATTRIBUTES:
				ar >> TempVal;
				m_TermsSelectedAttributes = (bool) (TempVal & 0x01);
				break;
			case es_SELECTIONRANDONNUMS:
				ar >> TempVal;
				m_TermsSelectedRandNums = (bool) (TempVal & 0x01);
				break;
			case es_SELECTIONNUMDECIMALS:
				ar >> m_TermsSelectedNumDecimals;
				break;
			case es_SELECTIONNUMLOW:
				ar >> m_TermsSelectedNumLow;
				break;
			case es_SELECTIONNUMHIGH:
				ar >> m_TermsSelectedNumHigh;
				break;
			case es_SELECTIONBOOLEANS:
				ar >> TempVal;
				m_TermsSelectedBools = (bool) (TempVal & 0x01);
				break;
			case es_SELECTIONSTRINGS:
				ar >> TempVal;
				m_TermsSelectedStrings = (bool) (TempVal & 0x01);
				break;
			case es_SELECTIONCHARSLOW:
				ar >> m_TermsSelectedCharLow;
				break;
			case es_SELECTIONCHARSHIGH:
				ar >> m_TermsSelectedCharHigh;
				break;
			case es_SELECTIONSTRLENLOW:
				ar >> m_TermsSelectedCharLenLow;
				break;
			case es_SELECTIONSTRLENHIGH:
				ar >> m_TermsSelectedCharLenHigh;
				break;
			case es_SELECTIONARGS:
				ar >> TempVal;
				m_TermsSelectedArgs = (bool) (TempVal & 0x01);
				break;
			case es_ENDENVIRONMENT:
				ContinueExec = FALSE;
				break;
			} // Command Switch 

		} // while buffer is not empty
		m_FileLoaded = TRUE;

	} // Loading (ar.IsStoring != TRUE)
} // Serialize

/////////////////////////////////////////////////////////////////////////////
// CEnvironmentDoc commands
void CEnvironmentDoc::SaveEnvironment()
{
	// Save Environment Method
	
	if (!m_FileLoaded)
	{
		CFile SaveEnvironFile;
		static char BASED_CODE szFilter[] = "GP-Lab Files (*.gps)|*.gps";

		// Create File Save As Dialog
		CFileDialog OurFileDlg (
			FALSE,                // Save As
			_T(".gps"),           // Default GAPS 2 Library File
			_T("Environment1"),       // Default GAPS 2 Library File Name
			OFN_OVERWRITEPROMPT,  // If file exists, ask if user wants to over-write
			szFilter);

		if (OurFileDlg.DoModal() == IDOK)
		{
			if (!SaveEnvironFile.Open(OurFileDlg.GetPathName(), CFile::modeCreate | CFile::modeWrite))
			{
				AfxMessageBox(CString("Could not write to file ") + OurFileDlg.GetPathName());
				return;
			}
			CArchive ArchiveLib(&SaveEnvironFile, CArchive::store);
			Serialize(ArchiveLib);
			ArchiveLib.Close();
			SaveEnvironFile.Close();
			m_FileName = OurFileDlg.GetPathName();
		}
	}
	else
	{
		if (!CDocument::OnSaveDocument(m_FileName))
			return;

		m_FileLoaded = TRUE;
	}
	
}

BOOL CEnvironmentDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	if (!CDocument::OnOpenDocument(lpszPathName))
		return FALSE;
	
	m_FileName = lpszPathName;
	m_FileLoaded = TRUE;
	
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// IsContextuallySound ensures that the document's selected functions, 
// all the arguments thereof, and selected built-ins will all be in synch
// with one another.  I.E., if a function requires the call of a boolean,
// at least one boolean function and/or terminal should always be available
// to select.
//////////////////////////////////////////////////////////////////////////////////
// Also, Check to ensure that if the RPB is describing statements, that the RPB 
// Context to be built is a VOID, and if the RPB is describing an non-VOID, that 
// the RPB Context to be built is not a VOID.  
bool CEnvironmentDoc::IsContextuallySound()
{
	int index;

	bool TFloatSelected     = FALSE;
	bool TBoolSelected      = FALSE;
	bool TStringSelected    = FALSE;

	bool StatementsChecked = FALSE;

	bool FloatNeeded = FALSE;
	bool BoolNeeded = FALSE;
	bool StringNeeded = FALSE;
	bool AttributesNeeded = FALSE;

	CString ThisFunction;
	int parmindex;
	enum RETURNTYPE ThisType;

	if (m_TermsSelectedArgs)
	{

		//////////////////////////////////////////
		// Go through Arguments
		if (tbl_OurTargets->Arg1Type == rtFLOAT)
			TFloatSelected = TRUE;
		if (tbl_OurTargets->Arg1Type == rtBOOL)
			TBoolSelected = TRUE;
		if (tbl_OurTargets->Arg1Type == rtSTRING)
			TStringSelected = TRUE;
		
		if (m_NumArguments > 1)
		{
			if (tbl_OurTargets->Arg2Type == rtFLOAT)
				TFloatSelected = TRUE;
			if (tbl_OurTargets->Arg2Type == rtBOOL)
				TBoolSelected = TRUE;
			if (tbl_OurTargets->Arg2Type == rtSTRING)
				TStringSelected = TRUE;
		}

		if (m_NumArguments > 2)
		{
			if (tbl_OurTargets->Arg3Type == rtFLOAT)
				TFloatSelected = TRUE;
			if (tbl_OurTargets->Arg3Type == rtBOOL)
				TBoolSelected = TRUE;
			if (tbl_OurTargets->Arg3Type == rtSTRING)
				TStringSelected = TRUE;
		}

		if (m_NumArguments > 3)
		{
			if (tbl_OurTargets->Arg4Type == rtFLOAT)
				TFloatSelected = TRUE;
			if (tbl_OurTargets->Arg4Type == rtBOOL)
				TBoolSelected = TRUE;
			if (tbl_OurTargets->Arg4Type == rtSTRING)
				TStringSelected = TRUE;
		}

		if (m_NumArguments > 4)
		{
			if (tbl_OurTargets->Arg5Type == rtFLOAT)
				TFloatSelected = TRUE;
			if (tbl_OurTargets->Arg5Type == rtBOOL)
				TBoolSelected = TRUE;
			if (tbl_OurTargets->Arg5Type == rtSTRING)
				TStringSelected = TRUE;
		}

		if (m_NumArguments > 5)
		{
			if (tbl_OurTargets->Arg6Type == rtFLOAT)
				TFloatSelected = TRUE;
			if (tbl_OurTargets->Arg6Type == rtBOOL)
				TBoolSelected = TRUE;
			if (tbl_OurTargets->Arg6Type == rtSTRING)
				TStringSelected = TRUE;
		}
	}

	if (m_TermsSelectedBools)
		TBoolSelected = TRUE;

	if (m_TermsSelectedRandNums)
		TFloatSelected = TRUE;

	if (m_TermsSelectedStrings)
		TStringSelected = TRUE;


	//////////////////////////////////////////
	// Go through Attributes
	int TotalSelectable = 0;
	if (m_TermsSelectedAttributes)
	{
		for (index = 0; index < m_TotalAttributes; index++)
		{
			if (!a_Attributes[index].IsSelectable)
				continue;
			if (a_Attributes[index].Type == CString("Number"))
				TFloatSelected = TRUE;
			if (a_Attributes[index].Type == CString("Boolean"))
				TBoolSelected = TRUE;
			if (a_Attributes[index].Type == CString("String"))
				TStringSelected = TRUE;
			TotalSelectable++;
		}

		if (TotalSelectable == 0)
		// If attributes are checked as being selectable, but no attributes are, this is not good.
		{
			AfxMessageBox ("You have checked the box that Attributes are selected, but there are no selectable attributes given.");
			return FALSE;
		}
	}
	

	for (index = 0; index < m_TotalSelectedBuiltIns; index++)
	{
		ThisFunction = a_SelectedBuiltIns[index].Mid(a_SelectedBuiltIns[index].Find(" ")+1);
		if (ThisFunction == CString("If") ||
			ThisFunction == CString("While") ||
			ThisFunction == CString("DoUntil") ||
			ThisFunction == CString("not") ||
			ThisFunction == CString("&&") ||
			ThisFunction == CString("||"))
			BoolNeeded = TRUE;

		if (ThisFunction == CString("+") ||
			ThisFunction == CString("-") ||
			ThisFunction == CString("/") ||
			ThisFunction == CString("*") ||
			ThisFunction == CString("%") ||
			ThisFunction == CString(">") ||
			ThisFunction == CString("<") ||
			ThisFunction == CString(">=") ||
			ThisFunction == CString("<=") ||
			ThisFunction == CString("sin") ||
			ThisFunction == CString("cos") ||
			ThisFunction == CString("tan") ||
			ThisFunction == CString("log") ||
			ThisFunction == CString("sqrt") ||
			ThisFunction == CString("round") ||
			ThisFunction == CString("trunc") ||
			ThisFunction == CString("abs")  ||
			ThisFunction == CString("Replace") ||
			ThisFunction == CString("GetAt"))
			FloatNeeded = TRUE;

		if (ThisFunction == CString("Replace") ||
		    ThisFunction == CString("GetAt"))
			StringNeeded = TRUE;

		if (ThisFunction == CString("If") ||
			ThisFunction == CString("While") ||
			ThisFunction == CString("DoUntil") ||
			ThisFunction == CString(":=") ||
			ThisFunction == CString("Push") ||
			ThisFunction == CString("Pop") ||
			ThisFunction == CString("Push") ||
			ThisFunction == CString("Replace"))
			StatementsChecked = TRUE;

		if (ThisFunction == CString(":=") ||
			ThisFunction == CString("Push") ||
			ThisFunction == CString("Pop") ||
			ThisFunction == CString("Push") ||
			ThisFunction == CString("Replace"))
			AttributesNeeded = TRUE;

	}

	if (StatementsChecked)
	{
		if (m_RPBContext == rtFLOAT ||
			 m_RPBContext == rtBOOL ||
			 m_RPBContext == rtSTRING ||
			 m_RPBContext == rtEXPRESSION)
		{
			AfxMessageBox ("Statements were selected, but the RPB() context is not a statement-generating branch.");
			return FALSE;
		}
		else
		{
			if (AttributesNeeded && (!m_TermsSelectedAttributes || TotalSelectable == 0)) 
			{
				AfxMessageBox ("Statement(s) that require attributes were selected, but the Select Attributes box is unchecked.");
				return FALSE;
			}
			return TRUE;
		}
	}

	for (index = 0; index < m_TotalUserEnteredFunctions; index++)
	{
		ThisFunction = tbl_OurFunctionTable.GetFunctionName(index);
		// Go through each of their parameters
		parmindex = 0;
		while (parmindex < tbl_OurFunctionTable.GetParameterCount(ThisFunction))
		{
			ThisType = tbl_OurFunctionTable.GetParmType(ThisFunction, parmindex);
			switch (ThisType)
			{
			case rtFLOAT:
				FloatNeeded = TRUE;
				break;
			case rtBOOL:
				BoolNeeded = TRUE;
				break;
			case rtSTRING:
				StringNeeded = TRUE;
				break;
			}
			parmindex++;
		}

		// Check function return type itself
		switch (tbl_OurFunctionTable.GetType(ThisFunction))
		{
		case rtFLOAT:
			TFloatSelected = TRUE;
			break;
		case rtBOOL:
			TBoolSelected = TRUE;
			break;
		case rtSTRING:
			TStringSelected = TRUE;
			break;
		}
	}

	if (TFloatSelected && !FloatNeeded ||
		!TFloatSelected && FloatNeeded)
		return FALSE;

	if (TBoolSelected && !BoolNeeded ||
		!TBoolSelected && BoolNeeded)
		return FALSE;

	if (TStringSelected && !StringNeeded ||
		!TStringSelected && StringNeeded)
		return FALSE;


	return TRUE;
	
}


