#include "CExecEnv.h"
#include "CGPProgram.h"


#if !defined(AFX_ENVIRONMENTDOC_H__2611F160_4746_4469_A39E_A8B6F5AC38AA__INCLUDED_)
#define AFX_ENVIRONMENTDOC_H__2611F160_4746_4469_A39E_A8B6F5AC38AA__INCLUDED_

#define MAXLIBRARIES 100

// Generation Strategies
#define gs_GENFULL 0
#define gs_GENGROW 1
#define gs_GENHALF 2

// Selection Strategies
#define ss_FITNESS 0
#define ss_TOURN   1
#define ss_OVERSEL 2

#define TOTALBUILTINS 100
#define TOTALMAXATTRIBUTES 100

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// EnvironmentDoc.h : header file
//

/////////////////////////////////////////////////////////////////////////////
// CEnvironmentDoc document

class CEnvironmentDoc : public CDocument
{
protected:
	CEnvironmentDoc();           // protected constructor used by dynamic creation
	DECLARE_DYNCREATE(CEnvironmentDoc)

// Attributes
public:

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CEnvironmentDoc)
	public:
	virtual void Serialize(CArchive& ar);   // overridden for document i/o
	virtual BOOL OnOpenDocument(LPCTSTR lpszPathName);
	protected:
	virtual BOOL OnNewDocument();
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CEnvironmentDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

	// Generated message map functions
protected:
	//{{AFX_MSG(CEnvironmentDoc)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

public:

	bool m_FileLoaded;
	CString m_FileName;
	
	// Go through the tabs (of the Environment Form):
	// 1) Environment Identification
	CString m_EnvironName;
	CString m_EnvironDescr;

	// 2) Environmental Parameters
	double  m_MutationRate;          // Percentage of crossovers/breedings that result in mutation
	int	    m_MaxUnproductiveGens;   // Total number of consecutive generations allowed in which
	                                 // no improvement is made
	int     m_ShrinkageGenerations;  // Number of post-termination generations used to attempt to find
	                                 // shorter - and hence, 'better' trees
	long    m_Population;            // Population size
	int	    m_HonorRollSize;         // N-best individuals listed after execution is complete
	long    m_AbsoluteTimeLimit;     // Total maximum number of crossovers and breeding
	double  m_FitnessThreshold;      // How 'close' is a numeric fitness match?
	int     m_GenerationStrategy;    // Which Generation strategy is being used
	int     m_MaximumDepth;          // Maximum Tree Depth
	int	    m_MinimumDepth;          // Minimum Tree Depth
	int     m_SelectionStrategy;     // Which selection strategy is being used.
	double  m_SelectionParmNum;      // For Overselection, %age of Group I selected
	int     m_TournamentSize;        // For Tournament-style, total # of candidates per selection

	// 3) Targets
	int     m_NumArguments;
	int     m_NumTargets;
	struct ArgsAndTargets *tbl_OurTargets; // Stores all Arguments and Targets

	// 4) User Functions
	struct UserEnteredFunction a_UserEnteredFunctions[100];
	int m_TotalUserEnteredFunctions;
	CString a_SelectedBuiltIns[TOTALBUILTINS];
	int m_TotalSelectedBuiltIns;

	// 5) Base Program
	CString m_BaseProgram;

	// 6) Fitness Calculation
	CString m_RawFitnessFormula;
	CString m_StandardizedFitnessFormula;

	// 7) Attributes and Selected Terminals
	struct UserAttributes a_Attributes[100];
	int    m_TotalAttributes;
	bool   m_TermsSelectedAttributes;
	bool   m_TermsSelectedRandNums;
	int    m_TermsSelectedNumDecimals;
	double m_TermsSelectedNumLow;
	double m_TermsSelectedNumHigh;
	bool   m_TermsSelectedBools;
	bool   m_TermsSelectedStrings;
	int    m_TermsSelectedCharLow;
	int    m_TermsSelectedCharHigh;
	int    m_TermsSelectedCharLenLow;
	int    m_TermsSelectedCharLenHigh;
	bool   m_TermsSelectedArgs;


	// Methods/Message Handlers
	void SaveEnvironment();
	void ResetTrees();
	bool IsContextuallySound();

	// Base Program Executable
	class CGPProgram * t_MainProgramTree;

	// User Function Table
	class CFunctionTable tbl_OurFunctionTable;

	// Function Selection Table
	class CFunctionSelector tbl_OurFunctionSelections;

	// Execution Tree of the Fitness Calculations
	class CGPProgram * t_RawFitnessTree;
	class CGPProgram * t_StdFitnessTree;

	// Result-Producing Branch Node Location
	class CGPProgram t_RPBNode;
	enum RETURNTYPE m_RPBContext;

};

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_ENVIRONMENTDOC_H__2611F160_4746_4469_A39E_A8B6F5AC38AA__INCLUDED_)
